import re
import string
from typing import Any
from unittest.mock import sentinel
from textwrap import dedent
from textwrap import indent
import pytest
from importlinter.application.output import console
from grimp import ImportGraph

from importlinter.application.app_config import settings
from importlinter.application.ports.building import GraphBuilder
from importlinter.application.use_cases import (
    FAILURE,
    SUCCESS,
    create_report,
    lint_imports,
)
from importlinter.application.user_options import UserOptions
from tests.adapters.building import FakeGraphBuilder
from tests.adapters.timing import FakeTimer
from tests.adapters.user_options import (
    ExceptionRaisingUserOptionReader,
    FakeUserOptionReader,
)
from importlinter.application import rendering

SOME_CACHE_DIR = "/path/to/some/cache/dir"

INDENTED_LOGO = indent(rendering.TEXT_LOGO, " " * 12)


class TestCheckContractsAndPrintReport:
    def test_all_successful(self):
        self._configure(
            contracts_options=[
                {"type": "always_passes", "name": "Contract foo"},
                {"type": "always_passes", "name": "Contract bar"},
            ]
        )

        with console.capture() as capture:
            result = lint_imports()

        assert result == SUCCESS
        assert capture.get() == dedent(
            f"""\
            {INDENTED_LOGO}
            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT
            Contract bar KEPT

            Contracts: 2 kept, 0 broken.
            """
        )

    def test_invalid_contract(self):
        self._configure(
            contracts_options=[
                {
                    "type": "fields",
                    "name": "Contract foo",
                    "single_field": ["one", "two"],
                    "multiple_field": "one",
                    "import_field": "foobar",
                },
                {"type": "always_passes", "name": "Contract bar"},
            ]
        )

        with console.capture() as capture:
            result = lint_imports()

        assert result == FAILURE
        assert capture.get() == dedent(
            f"""\
            {INDENTED_LOGO}
            Contract "Contract foo" is not configured correctly:
                single_field: Expected a single value, got multiple values.
                import_field: Must be in the form "package.importer -> package.imported".
                required_field: This is a required field.
            """
        )

    def test_one_failure(self):
        self._configure(
            contracts_options=[
                {"type": "always_fails", "name": "Contract foo"},
                {"type": "always_passes", "name": "Contract bar"},
            ]
        )

        with console.capture() as capture:
            result = lint_imports()

        assert result == FAILURE
        assert capture.get() == dedent(
            f"""\
            {INDENTED_LOGO}
            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo BROKEN
            Contract bar KEPT

            Contracts: 1 kept, 1 broken.


            ----------------
            Broken contracts
            ----------------

            Contract foo
            ------------

            This contract will always fail.
            """
        )

    def test_limit_contracts(self):
        self._configure(
            contracts_options=[
                {"type": "always_passes", "id": "blue", "name": "Contract blue"},
                {"type": "always_passes", "id": "green", "name": "Contract green"},
                {"type": "always_fails", "id": "yellow", "name": "Contract yellow"},
                {"type": "always_passes", "id": "purple", "name": "Contract purple"},
            ]
        )

        with console.capture() as capture:
            result = lint_imports(limit_to_contracts=("green", "purple"))

        assert result == SUCCESS
        assert capture.get() == dedent(
            f"""\
            {INDENTED_LOGO}
            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract green KEPT
            Contract purple KEPT

            Contracts: 2 kept, 0 broken.
            """
        )

    def test_warnings(self):
        self._configure(
            contracts_options=[
                {
                    "type": "always_passes",
                    "name": "Contract foo",
                    "warnings": ["Some warning.", "Another warning."],
                },
                {
                    "type": "always_fails",
                    "name": "Contract bar",
                    "warnings": ["A third warning."],
                },
            ]
        )

        with console.capture() as capture:
            result = lint_imports()

        assert result == FAILURE
        assert capture.get() == dedent(
            f"""\
            {INDENTED_LOGO}
            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT (2 warnings)
            Contract bar BROKEN (1 warning)

            Contracts: 1 kept, 1 broken.

            --------
            Warnings
            --------

            Contract foo
            ------------

            - Some warning.
            - Another warning.

            Contract bar
            ------------

            - A third warning.


            ----------------
            Broken contracts
            ----------------

            Contract bar
            ------------

            This contract will always fail.
            """
        )

    def test_timings(self):
        timer = FakeTimer()
        timer.setup(tick_duration=5, increment=10)
        self._configure(
            contracts_options=[
                {
                    "type": "always_passes",
                    "name": "Contract foo",
                },
                {
                    "type": "always_passes",
                    "name": "Contract bar",
                    "warnings": ["A warning."],
                },
            ],
            timer=timer,
        )

        with console.capture() as capture:
            lint_imports(show_timings=True)

        assert capture.get() == dedent(
            f"""\
            {INDENTED_LOGO}
            Building graph took 5.0s.

            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT [15s]
            Contract bar KEPT (1 warning) [25s]

            Contracts: 2 kept, 0 broken.

            --------
            Warnings
            --------

            Contract bar
            ------------

            - A warning.
            """
        )

    @pytest.mark.parametrize(
        "cache_dir, expected_graph_building_output",
        (
            (
                "/path/to/cache",
                "Building import graph (cache directory is /path/to/cache)...",
            ),
            (None, "Building import graph (with caching disabled)..."),
            (
                sentinel.not_supplied,
                f"Building import graph (cache directory is {SOME_CACHE_DIR})...",
            ),
        ),
    )
    @pytest.mark.parametrize(
        "verbose, expected_output_template",
        [
            (
                True,
                f"""\
            {INDENTED_LOGO}
            Verbose mode.
            << graph building output >>
            Built graph in 5.0s.
            Checking Contract foo...
            Hello from the noisy contract!
            Contract foo KEPT [15s]
            Checking Contract bar...
            Contract bar KEPT [25s]

            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT
            Contract bar KEPT

            Contracts: 2 kept, 0 broken.
                """,
            ),
            (
                False,
                f"""\
            {INDENTED_LOGO}
            ---------
            Contracts
            ---------

            Analyzed 26 files, 10 dependencies.
            -----------------------------------

            Contract foo KEPT
            Contract bar KEPT

            Contracts: 2 kept, 0 broken.
                """,
            ),
        ],
    )
    def test_verbose_mode(
        self,
        verbose,
        expected_output_template,
        cache_dir,
        expected_graph_building_output,
    ):
        timer = FakeTimer()
        timer.setup(tick_duration=5, increment=10)
        self._configure(
            contracts_options=[
                {"type": "noisy", "name": "Contract foo"},
                {"type": "always_passes", "name": "Contract bar"},
            ],
            timer=timer,
        )

        kwargs = dict(verbose=verbose, is_debug_mode=True)
        if cache_dir is not sentinel.not_supplied:
            kwargs["cache_dir"] = cache_dir

        with console.capture() as capture:
            lint_imports(**kwargs)

        expected_output = expected_output_template.replace(
            "<< graph building output >>", expected_graph_building_output
        )
        assert capture.get() == dedent(expected_output)

    @pytest.mark.parametrize(
        "passed_arg, expected_cache_dir",
        (
            (sentinel.not_supplied, SOME_CACHE_DIR),
            ("/path/to/cache", "/path/to/cache"),
            (None, None),
        ),
    )
    def test_cache_dir_is_passed_correctly(self, passed_arg, expected_cache_dir):
        builder = FakeGraphBuilder()
        self._configure(contracts_options=[], graph_builder=builder)

        if passed_arg is sentinel.not_supplied:
            lint_imports()
        else:
            lint_imports(cache_dir=passed_arg)

        assert builder.build_arguments["cache_dir"] == expected_cache_dir

    def test_forbidden_import(self):
        """
        Tests the ForbiddenImportContract - a simple contract that
        looks at the graph.
        """
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.foo",
            imported="mypackage.bar",
            line_number=8,
            line_contents="from mypackage import bar",
        )
        graph.add_import(
            importer="mypackage.foo",
            imported="mypackage.bar",
            line_number=16,
            line_contents="from mypackage.bar import something",
        )
        self._configure(
            contracts_options=[
                {"type": "always_passes", "name": "Contract foo"},
                {
                    "type": "forbidden",
                    "name": "Forbidden contract one",
                    "importer": "mypackage.foo",
                    "imported": "mypackage.bar",
                },
                {
                    "type": "forbidden",
                    "name": "Forbidden contract two",
                    "importer": "mypackage.foo",
                    "imported": "mypackage.baz",
                },
            ],
            graph=graph,
        )

        with console.capture() as capture:
            result = lint_imports()

        assert result == FAILURE

        # Expecting 28 files (default graph has 26 modules, we add 2).
        # Expecting 11 dependencies (default graph has 10 imports, we add 2,
        # but it counts as 1 as it's between the same modules).
        assert capture.get() == dedent(
            f"""\
            {INDENTED_LOGO}
            ---------
            Contracts
            ---------

            Analyzed 28 files, 11 dependencies.
            -----------------------------------

            Contract foo KEPT
            Forbidden contract one BROKEN
            Forbidden contract two KEPT

            Contracts: 2 kept, 1 broken.


            ----------------
            Broken contracts
            ----------------

            Forbidden contract one
            ----------------------

            mypackage.foo is not allowed to import mypackage.bar:

                mypackage.foo:8: from mypackage import bar
                mypackage.foo:16: from mypackage.bar import something
            """
        )

    def test_debug_mode_doesnt_swallow_exception(self):
        some_exception = RuntimeError("There was some sort of exception.")
        reader = ExceptionRaisingUserOptionReader(exception=some_exception)
        settings.configure(
            USER_OPTION_READERS={"foo": reader},
            GRAPH_BUILDER=FakeGraphBuilder(),
        )

        with pytest.raises(some_exception.__class__, match=str(some_exception)):
            lint_imports(is_debug_mode=True)

    def test_non_debug_mode_prints_exception(self):
        some_exception = RuntimeError("There was some sort of exception.")
        reader = ExceptionRaisingUserOptionReader(exception=some_exception)
        settings.configure(
            USER_OPTION_READERS={"foo": reader},
            GRAPH_BUILDER=FakeGraphBuilder(),
        )

        with console.capture() as capture:
            lint_imports(is_debug_mode=False)

        assert capture.get() == dedent(
            f"""\
            {INDENTED_LOGO}
            There was some sort of exception.
            """
        )

    def _configure(
        self,
        contracts_options: list[dict[str, Any]],
        session_options: dict[str, Any] | None = None,
        contract_types: list[str] | None = None,
        graph: ImportGraph | None = None,
        graph_builder: GraphBuilder | None = None,
        timer: FakeTimer | None = None,
    ):
        session_options = session_options or {"root_package": "mypackage"}
        if not contract_types:
            contract_types = [
                "always_passes: tests.helpers.contracts.AlwaysPassesContract",
                "always_fails: tests.helpers.contracts.AlwaysFailsContract",
                "fields: tests.helpers.contracts.FieldsContract",
                "forbidden: tests.helpers.contracts.ForbiddenImportContract",
                "noisy: tests.helpers.contracts.NoisyContract",
            ]
        session_options["contract_types"] = contract_types

        reader = FakeUserOptionReader(
            UserOptions(session_options=session_options, contracts_options=contracts_options)
        )
        settings.configure(
            USER_OPTION_READERS={"foo": reader},
            GRAPH_BUILDER=graph_builder or FakeGraphBuilder(),
            TIMER=timer or FakeTimer(),
            DEFAULT_CACHE_DIR=SOME_CACHE_DIR,
        )
        if graph is None:
            graph = self._build_default_graph()

        settings.GRAPH_BUILDER.inject_graph(graph)

    def _build_default_graph(self):
        graph = ImportGraph()

        # Add 26 modules.
        for letter in string.ascii_lowercase:
            graph.add_module(f"mypackage.{letter}")

        # Add 10 imports in total.
        for imported in ("d", "e", "f"):
            for importer in ("a", "b", "c"):
                graph.add_import(
                    importer=f"mypackage.{importer}", imported=f"mypackage.{imported}"
                )  # 3 * 3 = 9 imports.
        graph.add_import(importer="mypackage.d", imported="mypackage.f")  # 1 extra import.
        return graph


class TestMultipleRootPackages:
    def test_builder_is_called_with_root_packages(self):
        builder = FakeGraphBuilder()
        root_package_names = ["mypackageone", "mypackagetwo"]
        settings.configure(GRAPH_BUILDER=builder)

        create_report(
            UserOptions(
                session_options={"root_packages": root_package_names},
                contracts_options=[],
            )
        )

        assert builder.build_arguments["root_package_names"] == root_package_names


class TestGraphCopying:
    def test_graph_can_be_mutated_without_affecting_other_contracts(self):
        # The MutationCheckContract checks that there are a certain number of modules and imports
        # in the graph, then adds one more module and one more import. We can check two such
        # contracts and the second one will fail, if the graph gets mutated by other contracts.
        session_options = {
            "root_package": "mypackage",
            "contract_types": ["mutation_check: tests.helpers.contracts.MutationCheckContract"],
        }

        reader = FakeUserOptionReader(
            UserOptions(
                session_options=session_options,
                contracts_options=[
                    {
                        "type": "mutation_check",
                        "name": "Contract one",
                        "number_of_modules": "5",
                        "number_of_imports": "2",
                    },
                    {
                        "type": "mutation_check",
                        "name": "Contract two",
                        "number_of_modules": "5",
                        "number_of_imports": "2",
                    },
                ],
            )
        )
        settings.configure(
            USER_OPTION_READERS={"foo": reader},
            GRAPH_BUILDER=FakeGraphBuilder(),
        )

        graph = ImportGraph()

        # Create a graph with five modules and two imports.
        for module in ("one", "two", "three", "four", "five"):
            graph.add_module(module)
        graph.add_import(importer="one", imported="two")
        graph.add_import(importer="one", imported="three")

        settings.GRAPH_BUILDER.inject_graph(graph)

        result = lint_imports(is_debug_mode=True)

        assert result == SUCCESS


class TestCreateReport:
    @pytest.mark.parametrize(
        "limit_to_contracts, expected_message",
        (
            (
                ("a-contract-that-exists", "a-contract-that-doesnt-exist"),
                "Could not find contract 'a-contract-that-doesnt-exist'.\n\n"
                "You asked to limit the check to that contract, but nothing exists with that id.",
            ),
            (
                (
                    "a-contract-that-exists",
                    "a-contract-that-doesnt-exist",
                    "another-contract-that-doesnt-exist",
                ),
                "Could not find the following contract ids: "
                "a-contract-that-doesnt-exist, another-contract-that-doesnt-exist.\n\n"
                "You asked to limit the check to those contracts, but there are no "
                "contracts with those ids.",
            ),
        ),
    )
    def test_raises_exception_if_limited_to_nonexistent_contract(
        self, limit_to_contracts, expected_message
    ):
        settings.configure(GRAPH_BUILDER=FakeGraphBuilder())

        with pytest.raises(
            ValueError,
            match=re.escape(expected_message),
        ):
            create_report(
                user_options=UserOptions(
                    session_options={"root_packages": ["mypackage"]},
                    contracts_options=[
                        {
                            "type": "always_passes",
                            "id": "a-contract-that-exists",
                            "name": "Contract one",
                        },
                    ],
                ),
                limit_to_contracts=limit_to_contracts,
            )


class TestReadUserOptions:
    @pytest.mark.parametrize("filename", [".importlinter", "setup.cfg", "foo", "foo.bar"])
    def test_default_behavior(self, filename):
        expected_error = RuntimeError("expected")

        foo_reader = ExceptionRaisingUserOptionReader(AssertionError)
        ini_reader = ExceptionRaisingUserOptionReader(expected_error)
        toml_reader = ExceptionRaisingUserOptionReader(AssertionError)

        settings.configure(
            USER_OPTION_READERS={
                "foo": foo_reader,
                "ini": ini_reader,
                "toml": toml_reader,
            },
        )
        with pytest.raises(RuntimeError, match="expected"):
            lint_imports(filename, is_debug_mode=True)

    @pytest.mark.parametrize("filename", ["pyproject.toml", "foo.toml"])
    def test_toml_file(self, filename):
        expected_error = RuntimeError("expected")

        foo_reader = ExceptionRaisingUserOptionReader(AssertionError)
        ini_reader = ExceptionRaisingUserOptionReader(AssertionError)
        toml_reader = ExceptionRaisingUserOptionReader(expected_error)

        settings.configure(
            USER_OPTION_READERS={
                "foo": foo_reader,
                "ini": ini_reader,
                "toml": toml_reader,
            },
        )
        with pytest.raises(RuntimeError, match="expected"):
            lint_imports(filename, is_debug_mode=True)



import time

from importlinter.adapters.timing import SystemClockTimer
from tests.adapters.timing import FakeTimer


class TestSystemClockTimer:
    def test_unnested(self):
        some_seconds = 2

        with SystemClockTimer() as timer:
            time.sleep(some_seconds)

        assert timer.duration_in_ms >= some_seconds * 1000

    def test_nested(self):
        timer = SystemClockTimer()

        some_seconds = 1
        with timer:
            with timer:
                time.sleep(some_seconds)
                with timer:
                    time.sleep(some_seconds)
                inner_duration = timer.duration_in_ms
            middle_duration = timer.duration_in_ms
        outer_duration = timer.duration_in_ms

        assert inner_duration >= some_seconds * 1000
        assert middle_duration >= inner_duration + some_seconds * 1000
        assert outer_duration >= middle_duration


class TestFakeTimer:
    def test_unnested(self):
        timer = FakeTimer()
        timer.setup(tick_duration=10, increment=3)

        for expected in (10, 13, 16):
            with timer:
                pass
            assert timer.duration_in_ms == expected * 1000

    def test_nested(self):
        timer = FakeTimer()
        timer.setup(tick_duration=10, increment=3)
        # Note that the time ticks each time the timer exits,
        # so the duration is cumulative.

        with timer:
            with timer:
                with timer:
                    pass
                assert timer.duration_in_ms == 10 * 1000
            assert timer.duration_in_ms == 23 * 1000
        assert timer.duration_in_ms == 39 * 1000



from pathlib import Path

import pytest

import importlinter
from importlinter import api
import importlinter.adapters
import importlinter.adapters.filesystem


@pytest.mark.parametrize(
    "config_directory, config_file, expected",
    (
        (
            "testpackage",
            "setup.cfg",
            dict(
                session_options={"root_packages": ["testpackage"]},
                contracts_options=[
                    {
                        "name": "Test independence contract",
                        "id": "test-independence",
                        "type": "independence",
                        "modules": [
                            "testpackage.high.blue",
                            "testpackage.high.green",
                        ],
                        "ignore_imports": [
                            "testpackage.utils -> testpackage.high.green",
                            "testpackage.*.blue.* -> testpackage.indirect.*",
                        ],
                    }
                ],
            ),
        ),
        (
            "multipleroots",
            ".multiplerootskeptcontract.ini",
            dict(
                session_options={"root_packages": ["rootpackageblue", "rootpackagegreen"]},
                contracts_options=[
                    {
                        "name": "Multiple roots kept contract",
                        "id": "multiple-roots",
                        "type": "forbidden",
                        "source_modules": "rootpackageblue.one.alpha",
                        "forbidden_modules": "rootpackagegreen.two",
                    }
                ],
            ),
        ),
    ),
)
def test_read_config(config_directory, config_file, expected):
    this_directory = Path(__file__).parent
    config_filename = str(this_directory / ".." / "assets" / config_directory / config_file)

    result = api.read_configuration(config_filename)

    assert result == expected


# TOML is always encoded in UTF-8, no matter the platform. However, open() selects the default
# encoding based on the platform, which means that on Windows, a description like
#     Sätze müssen unabhängig von Blöcken sein
# shows up as
#     SÃ¤tze mÃ¼ssen unabhÃ¤ngig von BlÃ¶cken sein
# in the CLI output.
def test_read_config_toml_utf8_on_windows(monkeypatch):
    default_encoding = "cp1252"
    this_directory = Path(__file__).parent
    config_directory = "testpackage"
    config_file = ".unicode.toml"
    config_filename = str(this_directory / ".." / "assets" / config_directory / config_file)

    # It would be more correct to mock locale.getencoding() here, but open() doesn't actually
    # care about its return value, despite the docs making it sound like that would be the case.
    def mock_read(self, file_name, encoding) -> str:
        with open(file_name, encoding=encoding or default_encoding) as file:
            return file.read()

    monkeypatch.setattr(importlinter.adapters.filesystem.FileSystem, "read", mock_read)

    config_result = api.read_configuration(config_filename)
    result = config_result["contracts_options"][0]["name"]

    expected = "Cönträct tö tést µnícðde pœrs€ng"

    assert result == expected



import re

import pytest
from grimp import DetailedImport
from grimp import ImportGraph

from importlinter.domain.helpers import (
    MissingImport,
    add_imports,
    import_expressions_to_imports,
    module_expression_to_modules,
    pop_import_expressions,
    pop_imports,
    resolve_import_expressions,
)
from importlinter.domain.imports import (
    DirectImport,
    ImportExpression,
    Module,
    ModuleExpression,
)


class TestPopImports:
    IMPORTS: list[DetailedImport] = [
        dict(
            importer="mypackage.blue",
            imported="mypackage.green",
            line_number=10,
            line_contents="blahblahblah",
        ),
        dict(
            importer="mypackage.green",
            imported="mypackage.blue",
            line_number=2,
            line_contents="blahblah",
        ),
        dict(
            importer="mypackage.green",
            imported="mypackage.yellow",
            line_number=1,
            line_contents="blah",
        ),
    ]

    def test_succeeds(self) -> None:
        graph = self._build_graph(imports=self.IMPORTS)
        imports_to_pop = self.IMPORTS[0:2]
        import_to_leave = self.IMPORTS[2]

        result = pop_imports(
            graph,
            [
                DirectImport(importer=Module(i["importer"]), imported=Module(i["imported"]))
                for i in imports_to_pop
            ],
        )

        assert result == imports_to_pop
        assert graph.direct_import_exists(
            importer=import_to_leave["importer"], imported=import_to_leave["imported"]
        )
        assert graph.count_imports() == 1

    def test_raises_missing_import_if_module_not_found(self) -> None:
        graph = self._build_graph(imports=self.IMPORTS)
        non_existent_import = DirectImport(
            importer=Module("mypackage.nonexistent"),
            imported=Module("mypackage.yellow"),
            line_number=1,
            line_contents="-",
        )
        with pytest.raises(
            MissingImport,
            match=re.escape(
                "Ignored import mypackage.nonexistent -> mypackage.yellow "
                "not present in the graph."
            ),
        ):
            pop_imports(graph, [non_existent_import])

    def test_works_with_multiple_external_imports_from_same_module(self) -> None:
        imports_to_pop: list[DetailedImport] = [
            dict(
                importer="mypackage.green",
                imported="someexternalpackage",
                line_number=2,
                line_contents="from someexternalpackage import one",
            ),
            dict(
                importer="mypackage.green",
                imported="someexternalpackage",
                line_number=2,
                line_contents="from someexternalpackage import two",
            ),
        ]
        imports = self.IMPORTS + imports_to_pop
        graph = self._build_graph(imports=imports)

        result = pop_imports(
            graph,
            [
                DirectImport(
                    importer=Module(i["importer"]),
                    imported=Module(i["imported"]),
                    line_number=i["line_number"],
                    line_contents=i["line_contents"],
                )
                for i in imports_to_pop
            ],
        )

        assert result == imports_to_pop
        one_of_the_popped_imports = imports_to_pop[0]
        assert not graph.direct_import_exists(
            importer=one_of_the_popped_imports["importer"],
            imported=one_of_the_popped_imports["imported"],
        )
        assert graph.count_imports() == len(self.IMPORTS)

    def _build_graph(self, imports):
        graph = ImportGraph()
        for import_ in imports:
            graph.add_import(**import_)
        return graph


class TestImportExpressionsToImports:
    DIRECT_IMPORTS = [
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.yellow"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.blue"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.blue"),
            imported=Module("mypackage.green"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.blue.cats"),
            imported=Module("mypackage.purple.dogs"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green.cats"),
            imported=Module("mypackage.orange.dogs"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green.cats"),
            imported=Module("mypackage.orange.mice"),
            line_number=1,
            line_contents="-",
        ),
        # Direct imports of external packages can appear more than once, as the external package
        # is squashed.
        DirectImport(
            importer=Module("mypackage.brown"),
            imported=Module("someotherpackage"),
            line_number=1,
            line_contents="from someotherpackage import one",
        ),
        DirectImport(
            importer=Module("mypackage.brown"),
            imported=Module("someotherpackage"),
            line_number=2,
            line_contents="from someotherpackage import two",
        ),
    ]

    @pytest.mark.parametrize(
        "description, expressions, expected",
        [
            (
                "No wildcards",
                [
                    ImportExpression(
                        importer=ModuleExpression(DIRECT_IMPORTS[0].importer.name),
                        imported=ModuleExpression(DIRECT_IMPORTS[0].imported.name),
                    ),
                ],
                [DIRECT_IMPORTS[0]],
            ),
            (
                "Importer wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.*"),
                        imported=ModuleExpression("mypackage.blue"),
                    ),
                ],
                [DIRECT_IMPORTS[1]],
            ),
            (
                "Imported wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green"),
                        imported=ModuleExpression("mypackage.*"),
                    ),
                ],
                DIRECT_IMPORTS[0:2],
            ),
            (
                "Importer and imported wildcards",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.*"),
                        imported=ModuleExpression("mypackage.*"),
                    ),
                ],
                DIRECT_IMPORTS[0:3],
            ),
            (
                "Inner wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.*.cats"),
                        imported=ModuleExpression("mypackage.*.dogs"),
                    ),
                ],
                DIRECT_IMPORTS[3:5],
            ),
            (
                "Multiple expressions, non-overlapping",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green"),
                        imported=ModuleExpression("mypackage.*"),
                    ),
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green.cats"),
                        imported=ModuleExpression("mypackage.orange.*"),
                    ),
                ],
                DIRECT_IMPORTS[0:2] + DIRECT_IMPORTS[4:6],
            ),
            (
                "Multiple expressions, overlapping",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.*"),
                        imported=ModuleExpression("mypackage.blue"),
                    ),
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green"),
                        imported=ModuleExpression("mypackage.blue"),
                    ),
                ],
                [DIRECT_IMPORTS[1]],
            ),
            (
                "Multiple imports of external package with same importer",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.brown"),
                        imported=ModuleExpression("someotherpackage"),
                    ),
                ],
                DIRECT_IMPORTS[6:8],
            ),
        ],
    )
    def test_succeeds(self, description, expressions, expected):
        graph = self._build_graph(self.DIRECT_IMPORTS)

        assert sorted(
            import_expressions_to_imports(graph, expressions),
            key=_direct_import_sort_key,
        ) == sorted(expected, key=_direct_import_sort_key)

    def test_raises_missing_import(self):
        graph = ImportGraph()
        graph.add_module("mypackage")
        graph.add_module("other")
        graph.add_import(
            importer="mypackage.b",
            imported="other.foo",
            line_number=1,
            line_contents="-",
        )

        expression = ImportExpression(
            importer=ModuleExpression("mypackage.a.*"),
            imported=ModuleExpression("other.foo"),
        )
        with pytest.raises(MissingImport):
            import_expressions_to_imports(graph, [expression])

    def _build_graph(self, direct_imports):
        graph = ImportGraph()
        for direct_import in direct_imports:
            graph.add_import(
                importer=direct_import.importer.name,
                imported=direct_import.imported.name,
                line_number=direct_import.line_number,
                line_contents=direct_import.line_contents,
            )
        return graph


class TestResolveImportExpressions:
    DIRECT_IMPORTS = [
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.yellow"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.blue"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.blue"),
            imported=Module("mypackage.green"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.blue.cats"),
            imported=Module("mypackage.purple.dogs"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green.cats"),
            imported=Module("mypackage.orange.dogs"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green.cats"),
            imported=Module("mypackage.orange.mice"),
            line_number=1,
            line_contents="-",
        ),
        # Direct imports of external packages can appear more than once, as the external package
        # is squashed.
        DirectImport(
            importer=Module("mypackage.brown"),
            imported=Module("someotherpackage"),
            line_number=1,
            line_contents="from someotherpackage import one",
        ),
        DirectImport(
            importer=Module("mypackage.brown"),
            imported=Module("someotherpackage"),
            line_number=2,
            line_contents="from someotherpackage import two",
        ),
        DirectImport(
            importer=Module("mypackage.green.cats.very.deep.module.cats"),
            imported=Module("mypackage.orange.mice.another.verydeep.one.dogs"),
            line_number=1,
            line_contents="-",
        ),
    ]

    @pytest.mark.parametrize(
        "description, expressions, expected_imports",
        [
            (
                "No wildcards",
                [
                    ImportExpression(
                        importer=ModuleExpression(DIRECT_IMPORTS[0].importer.name),
                        imported=ModuleExpression(DIRECT_IMPORTS[0].imported.name),
                    ),
                ],
                {DIRECT_IMPORTS[0]},
            ),
            (
                "Importer wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.*"),
                        imported=ModuleExpression("mypackage.blue"),
                    ),
                ],
                {DIRECT_IMPORTS[1]},
            ),
            (
                "Imported wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green"),
                        imported=ModuleExpression("mypackage.*"),
                    ),
                ],
                set(DIRECT_IMPORTS[0:2]),
            ),
            (
                "Importer and imported wildcards",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.*"),
                        imported=ModuleExpression("mypackage.*"),
                    ),
                ],
                set(DIRECT_IMPORTS[0:3]),
            ),
            (
                "Inner wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.*.cats"),
                        imported=ModuleExpression("mypackage.*.dogs"),
                    ),
                ],
                set(DIRECT_IMPORTS[3:5]),
            ),
            (
                "Importer recursive wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.**"),
                        imported=ModuleExpression("mypackage.blue"),
                    ),
                ],
                {DIRECT_IMPORTS[1]},
            ),
            (
                "Imported recursive wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green"),
                        imported=ModuleExpression("mypackage.**"),
                    ),
                ],
                set(DIRECT_IMPORTS[0:2]),
            ),
            (
                "Importer and imported recursive wildcards",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.**"),
                        imported=ModuleExpression("mypackage.**"),
                    ),
                ],
                set(DIRECT_IMPORTS[0:6]) | {DIRECT_IMPORTS[8]},
            ),
            (
                "Inner recursive wildcard",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.**.cats"),
                        imported=ModuleExpression("mypackage.**.dogs"),
                    ),
                ],
                set(DIRECT_IMPORTS[3:5]) | {DIRECT_IMPORTS[8]},
            ),
            (
                "Multiple expressions, non-overlapping",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green"),
                        imported=ModuleExpression("mypackage.*"),
                    ),
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green.cats"),
                        imported=ModuleExpression("mypackage.orange.*"),
                    ),
                ],
                set(DIRECT_IMPORTS[0:2] + DIRECT_IMPORTS[4:6]),
            ),
            (
                "Multiple expressions, overlapping",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.*"),
                        imported=ModuleExpression("mypackage.blue"),
                    ),
                    ImportExpression(
                        importer=ModuleExpression("mypackage.green"),
                        imported=ModuleExpression("mypackage.blue"),
                    ),
                ],
                {DIRECT_IMPORTS[1]},
            ),
            (
                "Multiple imports of external package with same importer",
                [
                    ImportExpression(
                        importer=ModuleExpression("mypackage.brown"),
                        imported=ModuleExpression("someotherpackage"),
                    ),
                ],
                set(DIRECT_IMPORTS[6:8]),
            ),
        ],
    )
    def test_succeeds(
        self,
        description: str,
        expressions: list[ImportExpression],
        expected_imports: list[DirectImport],
    ):
        graph = self._build_graph(self.DIRECT_IMPORTS)

        imports, unresolved_expressions = resolve_import_expressions(graph, expressions)

        assert unresolved_expressions == set()
        assert imports == expected_imports

    def test_detects_unresolved_expression(self):
        graph = ImportGraph()
        graph.add_module("mypackage")
        graph.add_module("other")
        graph.add_import(
            importer="mypackage.b",
            imported="other.foo",
            line_number=1,
            line_contents="-",
        )
        expression = ImportExpression(
            importer=ModuleExpression("mypackage.a.*"),
            imported=ModuleExpression("other.foo"),
        )

        imports, unresolved_expressions = resolve_import_expressions(graph, [expression])

        assert (imports, unresolved_expressions) == (
            set(),
            {
                ImportExpression(
                    imported=ModuleExpression("other.foo"),
                    importer=ModuleExpression("mypackage.a.*"),
                )
            },
        )

    def _build_graph(self, direct_imports):
        graph = ImportGraph()
        for direct_import in direct_imports:
            graph.add_import(
                importer=direct_import.importer.name,
                imported=direct_import.imported.name,
                line_number=direct_import.line_number,
                line_contents=direct_import.line_contents,
            )
        return graph


class TestPopImportExpressions:
    DIRECT_IMPORTS = [
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.yellow"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.blue"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.blue"),
            imported=Module("mypackage.green"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.blue.cats"),
            imported=Module("mypackage.purple.dogs"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green.cats"),
            imported=Module("mypackage.orange.dogs"),
            line_number=1,
            line_contents="-",
        ),
    ]

    def test_succeeds(self) -> None:
        graph = self._build_graph(self.DIRECT_IMPORTS)
        expressions = [
            ImportExpression(
                importer=ModuleExpression("mypackage.green"),
                imported=ModuleExpression("mypackage.*"),
            ),
            # Expressions can overlap.
            ImportExpression(
                importer=ModuleExpression("mypackage.green"),
                imported=ModuleExpression("mypackage.blue"),
            ),
            ImportExpression(
                importer=ModuleExpression("mypackage.blue.cats"),
                imported=ModuleExpression("mypackage.purple.dogs"),
            ),
        ]

        popped_imports: list[DetailedImport] = pop_import_expressions(graph, expressions)

        # Cast to direct imports to make comparison easier.
        popped_direct_imports: list[DirectImport] = sorted(
            map(self._dict_to_direct_import, popped_imports),
            key=_direct_import_sort_key,
        )
        expected = sorted(
            [
                self.DIRECT_IMPORTS[0],
                self.DIRECT_IMPORTS[1],
                self.DIRECT_IMPORTS[3],
            ],
            key=_direct_import_sort_key,
        )
        assert popped_direct_imports == expected
        assert graph.count_imports() == 2

    def _build_graph(self, direct_imports):
        graph = ImportGraph()
        for direct_import in direct_imports:
            graph.add_import(
                importer=direct_import.importer.name,
                imported=direct_import.imported.name,
                line_number=direct_import.line_number,
                line_contents=direct_import.line_contents,
            )
        return graph

    def _dict_to_direct_import(self, import_details: DetailedImport) -> DirectImport:
        return DirectImport(
            importer=Module(import_details["importer"]),
            imported=Module(import_details["imported"]),
            line_number=import_details["line_number"],
            line_contents=import_details["line_contents"],
        )


class TestModuleExpressionToModules:
    def _build_default_graph(self) -> ImportGraph:
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.bar",
            "mypackage.bar.one",
            "mypackage.bar.one.red",
            "mypackage.bar.one.blue",
            "mypackage.bar.one.green",
            "mypackage.bar.two",
            "mypackage.bar.two.red",
            "mypackage.bar.two.blue",
            "mypackage.bar.two.green",
            "mypackage.bar.three",
            "mypackage.bar.three.red",
            "mypackage.bar.three.blue",
            "mypackage.bar.three.green",
            "mypackage.foo",
            "mypackage.foo.one",
            "mypackage.foo.one.red",
            "mypackage.foo.one.blue",
            "mypackage.foo.one.green",
            "mypackage.toto",
            "mypackage.toto.red",
            "mypackage.toto.blue",
            "mypackage.toto.green",
        ):
            graph.add_module(module)
        return graph

    @pytest.mark.parametrize(
        "expression,expected,description",
        [
            (
                "mypackage.foo.**",
                {
                    "mypackage.foo.one",
                    "mypackage.foo.one.red",
                    "mypackage.foo.one.blue",
                    "mypackage.foo.one.green",
                },
                "Double wildcard at the end of expression",
            ),
            (
                "mypackage.**.red",
                {
                    "mypackage.bar.one.red",
                    "mypackage.bar.two.red",
                    "mypackage.bar.three.red",
                    "mypackage.foo.one.red",
                    "mypackage.toto.red",
                },
                "Double wildcard in the middle of expression",
            ),
            (
                "mypackage.**.one",
                {
                    "mypackage.bar.one",
                    "mypackage.foo.one",
                },
                "Double wildcard in the middle of expression where expected target has submodules",
            ),
            (
                "mypackage.bar.*",
                {
                    "mypackage.bar.one",
                    "mypackage.bar.two",
                    "mypackage.bar.three",
                },
                "Simple wildcard at the end of expression",
            ),
            (
                "mypackage.bar.*.red",
                {
                    "mypackage.bar.one.red",
                    "mypackage.bar.two.red",
                    "mypackage.bar.three.red",
                },
                "Simple wildcard in the middle of expression",
            ),
            (
                "mypackage.*.one",
                {
                    "mypackage.bar.one",
                    "mypackage.foo.one",
                },
                "Simple wildcard in the middle of expression where expected target has submodules",
            ),
            (
                "mypackage.bar.one",
                {
                    "mypackage.bar.one",
                },
                "No wildcard",
            ),
        ],
    )
    def test_expected_conversion(self, expression: str, expected: set[str], description: str):
        graph = self._build_default_graph()

        conversion_result = module_expression_to_modules(
            graph, expression=ModuleExpression(expression)
        )
        expected_modules = set(map(lambda name: Module(name), expected))

        assert conversion_result == expected_modules, description

    @pytest.mark.parametrize(
        "expression,expected,description",
        [
            (
                "mypackage.foo.**",
                {
                    "mypackage.foo.one",
                    "mypackage.foo.one.red",
                    "mypackage.foo.one.blue",
                    "mypackage.foo.one.green",
                },
                "Double wildcard at the end of expression",
            ),
            (
                "mypackage.**.red",
                {
                    "mypackage.bar.one.red",
                    "mypackage.bar.two.red",
                    "mypackage.bar.three.red",
                    "mypackage.foo.one.red",
                    "mypackage.toto.red",
                },
                "Double wildcard in the middle of expression",
            ),
            (
                "mypackage.**.one",
                {
                    "mypackage.bar.one",
                    "mypackage.bar.one.red",
                    "mypackage.bar.one.blue",
                    "mypackage.bar.one.green",
                    "mypackage.foo.one",
                    "mypackage.foo.one.red",
                    "mypackage.foo.one.blue",
                    "mypackage.foo.one.green",
                },
                "Double wildcard in the middle of expression where expected target has submodules",
            ),
            (
                "mypackage.bar.*",
                {
                    "mypackage.bar.one",
                    "mypackage.bar.one.red",
                    "mypackage.bar.one.blue",
                    "mypackage.bar.one.green",
                    "mypackage.bar.two",
                    "mypackage.bar.two.red",
                    "mypackage.bar.two.blue",
                    "mypackage.bar.two.green",
                    "mypackage.bar.three",
                    "mypackage.bar.three.red",
                    "mypackage.bar.three.blue",
                    "mypackage.bar.three.green",
                },
                "Simple wildcard at the end of expression",
            ),
            (
                "mypackage.bar.*.red",
                {
                    "mypackage.bar.one.red",
                    "mypackage.bar.two.red",
                    "mypackage.bar.three.red",
                },
                "Simple wildcard in the middle of expression",
            ),
            (
                "mypackage.*.one",
                {
                    "mypackage.bar.one",
                    "mypackage.bar.one.red",
                    "mypackage.bar.one.blue",
                    "mypackage.bar.one.green",
                    "mypackage.foo.one",
                    "mypackage.foo.one.red",
                    "mypackage.foo.one.blue",
                    "mypackage.foo.one.green",
                },
                "Simple wildcard in the middle of expression where expected target has submodules",
            ),
            (
                "mypackage.bar.one",
                {
                    "mypackage.bar.one",
                    "mypackage.bar.one.red",
                    "mypackage.bar.one.blue",
                    "mypackage.bar.one.green",
                },
                "No wildcard",
            ),
        ],
    )
    def test_expected_conversion_with_as_package_option(
        self, expression: str, expected: set[str], description: str
    ):
        graph = self._build_default_graph()

        conversion_result = module_expression_to_modules(
            graph, expression=ModuleExpression(expression), as_packages=True
        )
        expected_modules = set(map(lambda name: Module(name), expected))

        assert conversion_result == expected_modules, description


def test_add_imports() -> None:
    graph = ImportGraph()
    import_details: list[DetailedImport] = [
        {
            "importer": "a",
            "imported": "b",
            "line_number": 1,
            "line_contents": "lorem ipsum",
        },
        {
            "importer": "c",
            "imported": "d",
            "line_number": 2,
            "line_contents": "lorem ipsum 2",
        },
    ]
    assert not graph.modules
    add_imports(graph, import_details)
    assert graph.modules == {"a", "b", "c", "d"}


def _direct_import_sort_key(
    direct_import: DirectImport,
) -> tuple[str, str, int | None]:
    # Doesn't matter how we sort, just a way of sorting consistently for comparison.
    return (
        direct_import.importer.name,
        direct_import.imported.name,
        direct_import.line_number,
    )



import pytest
from textwrap import dedent
from grimp import ImportGraph
from importlinter.application.output import console
from importlinter.application.app_config import settings
from importlinter.contracts.layers import Layer, LayerField, LayersContract, ModuleTail
from importlinter.domain.contract import ContractCheck, InvalidContractOptions
from importlinter.domain.helpers import MissingImport
from importlinter.domain import fields
from tests.adapters.timing import FakeTimer


@pytest.fixture(scope="module", autouse=True)
def configure():
    settings.configure(TIMER=FakeTimer())


@pytest.mark.parametrize(
    "data, parsed_data",
    (
        ("one", Layer({ModuleTail(name="one", is_optional=False)})),
        ("(one)", Layer({ModuleTail(name="one", is_optional=True)})),
        (
            "one | two |    three",
            Layer(
                {
                    ModuleTail(name="one"),
                    ModuleTail(name="two"),
                    ModuleTail(name="three"),
                }
            ),
        ),
        (
            "one | (two) | three",
            Layer(
                {
                    ModuleTail(name="one"),
                    ModuleTail(name="two", is_optional=True),
                    ModuleTail(name="three"),
                }
            ),
        ),
        (
            "one : two :    three",
            Layer(
                {
                    ModuleTail(name="one"),
                    ModuleTail(name="two"),
                    ModuleTail(name="three"),
                },
                is_independent=False,
            ),
        ),
        (
            "one : (two) : three",
            Layer(
                {
                    ModuleTail(name="one"),
                    ModuleTail(name="two", is_optional=True),
                    ModuleTail(name="three"),
                },
                is_independent=False,
            ),
        ),
    ),
)
def test_layer_field(data, parsed_data):
    assert LayerField().parse(data) == parsed_data


def test_layer_field_raises_error_if_both_independent_and_non_independent():
    with pytest.raises(fields.ValidationError) as exc_info:
        LayerField().parse("one | two : three")
    assert (
        exc_info.value.message
        == "Layer cannot have a mixture of independent and non-independent elements."
    )


class TestLayerContractSingleContainers:
    def test_no_illegal_imports_means_contract_is_kept(self):
        contract = self._build_contract()
        graph = self._build_graph()

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is True

    def test_illegal_child_imports_means_contract_is_broken(self):
        contract = self._build_contract()
        graph = self._build_graph()
        graph.add_import(importer="mypackage.medium.orange", imported="mypackage.high.green")

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False

    def test_illegal_grandchild_to_child_means_contract_is_broken(self):
        contract = self._build_contract()
        graph = self._build_graph()
        graph.add_import(importer="mypackage.low.white.gamma", imported="mypackage.medium.red")

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False

    def _build_graph(self):
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.high",
            "mypackage.high.green",
            "mypackage.high.blue",
            "mypackage.high.yellow",
            "mypackage.high.yellow.alpha",
            "mypackage.medium",
            "mypackage.medium.orange",
            "mypackage.medium.orange.beta",
            "mypackage.medium.red",
            "mypackage.low",
            "mypackage.low.black",
            "mypackage.low.white",
            "mypackage.low.white.gamma",
        ):
            graph.add_module(module)

        # Add some 'legal' imports.
        graph.add_import(importer="mypackage.high.green", imported="mypackage.medium.orange")
        graph.add_import(importer="mypackage.high.green", imported="mypackage.low.white.gamma")
        graph.add_import(importer="mypackage.medium.orange", imported="mypackage.low.white")
        graph.add_import(importer="mypackage.high.blue", imported="mypackage.utils")
        graph.add_import(importer="mypackage.utils", imported="mypackage.medium.red")

        return graph

    def _build_contract(self):
        return LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "containers": ["mypackage"],
                "layers": ["high", "medium", "low"],
            },
        )


class TestLayerContractSiblingLayers:
    @pytest.mark.parametrize("layer_independent", (True, False))
    @pytest.mark.parametrize("specify_container", (True, False))
    def test_imports_between_sibling_modules(
        self,
        specify_container: bool,
        layer_independent: bool,
    ):
        contract = self._create_contract(
            specify_container=specify_container,
            layer_independent=layer_independent,
        )
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.high",
            "mypackage.medium_a",
            "mypackage.medium_b",
            "mypackage.medium_c",
            "mypackage.low",
        ):
            graph.add_module(module)
        # Add some legal imports.
        graph.add_import(importer="mypackage.high.green", imported="mypackage.medium.orange")
        graph.add_import(importer="mypackage.utils", imported="mypackage.medium.red")
        # Add an import between sibling modules within the medium layer.
        graph.add_import(
            importer="mypackage.medium_a.blue",
            imported="mypackage.medium_b.red",
            line_number=3,
            line_contents="-",
        )

        contract_check = contract.check(graph=graph, verbose=False)

        if layer_independent:
            assert contract_check.kept is False
            sorted_metadata = _get_sorted_metadata(contract_check)
            assert sorted_metadata == {
                "invalid_dependencies": [
                    {
                        "importer": "mypackage.medium_a",
                        "imported": "mypackage.medium_b",
                        "routes": [
                            {
                                "chain": [
                                    {
                                        "importer": "mypackage.medium_a.blue",
                                        "imported": "mypackage.medium_b.red",
                                        "line_numbers": (3,),
                                    },
                                ],
                                "extra_firsts": [],
                                "extra_lasts": [],
                            }
                        ],
                    },
                ],
                "undeclared_modules": set(),
            }
        else:
            assert contract_check.kept is True

    def _create_contract(self, specify_container: bool, layer_independent: bool):
        package = "mypackage"
        layer_prefix = "" if specify_container else f"{package}."
        layer_delimiter = "|" if layer_independent else ":"
        contract_options = {
            "layers": [
                f"{layer_prefix}high",
                f"{layer_prefix}medium_a {layer_delimiter} {layer_prefix}medium_b "
                f"{layer_delimiter} {layer_prefix}medium_c",
                f"{layer_prefix}low",
            ]
        }
        if specify_container:
            contract_options["containers"] = [package]
        return LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options=contract_options,
        )


class TestLayerMultipleContainers:
    def test_no_illegal_imports_means_contract_is_kept(self):
        contract = self._build_contract()
        graph = self._build_graph()

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is True

    def test_imports_from_low_to_high_but_in_different_container_doesnt_break_contract(
        self,
    ):
        contract = self._build_contract()
        graph = self._build_graph()
        graph.add_import(
            importer="mypackage.two.medium.green.beta",
            imported="mypackage.one.high.green",
        )
        graph.add_import(
            importer="mypackage.three.low.cyan", imported="mypackage.two.high.red.alpha"
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is True

    def test_illegal_grandchild_imports_means_contract_is_broken(self):
        contract = self._build_contract()
        graph = self._build_graph()
        graph.add_import(
            importer="mypackage.two.medium.green.beta",
            imported="mypackage.two.high.red.alpha",
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False

    def test_import_via_noncontainer_means_contract_is_broken(self):
        contract = self._build_contract()
        graph = self._build_graph()
        graph.add_import(
            importer="mypackage.one.medium.orange",
            imported="mypackage.noncontainer.blue",
        )
        graph.add_import(
            importer="mypackage.noncontainer.blue",
            imported="mypackage.two.high.red.alpha",
        )
        graph.add_import(
            importer="mypackage.two.high.red.alpha", imported="mypackage.one.high.green"
        )
        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False

    def _build_graph(self):
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.one",
            "mypackage.one.high",
            "mypackage.one.high.green",
            "mypackage.one.high.blue",
            "mypackage.one.high.yellow",
            "mypackage.one.high.yellow.alpha",
            "mypackage.one.medium",
            "mypackage.one.medium.orange",
            "mypackage.one.medium.orange.beta",
            "mypackage.one.medium.red",
            "mypackage.one.low",
            "mypackage.one.low.black",
            "mypackage.one.low.white",
            "mypackage.one.low.white.gamma",
            "mypackage.two",
            "mypackage.two.high",
            "mypackage.two.high.red",
            "mypackage.two.high.red.alpha",
            "mypackage.two.medium",
            "mypackage.two.medium.green",
            "mypackage.two.medium.green.beta",
            "mypackage.two.low",
            "mypackage.two.low.blue",
            "mypackage.two.low.blue.gamma",
            "mypackage.three",
            "mypackage.three.high",
            "mypackage.three.high.white",
            "mypackage.three.medium",
            "mypackage.three.medium.purple",
            "mypackage.three.low",
            "mypackage.three.low.cyan",
            "mypackage.noncontainer",
            "mypackage.noncontainer.blue",
        ):
            graph.add_module(module)

        # Add some 'legal' imports, each within their separate containers.
        graph.add_import(
            importer="mypackage.one.high.green", imported="mypackage.one.medium.orange"
        )
        graph.add_import(
            importer="mypackage.one.high.green",
            imported="mypackage.one.low.white.gamma",
        )
        graph.add_import(
            importer="mypackage.one.medium.orange", imported="mypackage.one.low.white"
        )

        graph.add_import(
            importer="mypackage.two.high.red.alpha",
            imported="mypackage.two.medium.green.beta",
        )
        graph.add_import(
            importer="mypackage.two.high.red.alpha",
            imported="mypackage.two.low.blue.gamma",
        )
        graph.add_import(
            importer="mypackage.two.medium.green.beta",
            imported="mypackage.two.low.blue.gamma",
        )

        graph.add_import(
            importer="mypackage.three.high.white",
            imported="mypackage.three.medium.purple",
        )
        graph.add_import(
            importer="mypackage.three.high.white", imported="mypackage.three.low.cyan"
        )
        graph.add_import(
            importer="mypackage.three.medium.purple",
            imported="mypackage.three.low.cyan",
        )

        return graph

    def _build_contract(self):
        return LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "containers": ["mypackage.one", "mypackage.two", "mypackage.three"],
                "layers": ["high", "medium", "low"],
            },
        )


class TestLayerContractWildcardContainers:
    def _build_graph(self):
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.components.one",
            "mypackage.components.one.high",
            "mypackage.components.one.high.green",
            "mypackage.components.one.high.blue",
            "mypackage.components.one.high.yellow",
            "mypackage.components.one.high.yellow.alpha",
            "mypackage.components.one.medium",
            "mypackage.components.one.medium.orange",
            "mypackage.components.one.medium.orange.beta",
            "mypackage.components.one.medium.red",
            "mypackage.components.one.low",
            "mypackage.components.one.low.black",
            "mypackage.components.one.low.white",
            "mypackage.components.one.low.white.gamma",
            "mypackage.components.two",
            "mypackage.components.two.high",
            "mypackage.components.two.high.red",
            "mypackage.components.two.high.red.alpha",
            "mypackage.components.two.medium",
            "mypackage.components.two.medium.green",
            "mypackage.components.two.medium.green.beta",
            "mypackage.components.two.low",
            "mypackage.components.two.low.blue",
            "mypackage.components.two.low.blue.gamma",
            "mypackage.components.three",
            "mypackage.components.three.high",
            "mypackage.components.three.high.white",
            "mypackage.components.three.medium",
            "mypackage.components.three.medium.purple",
            "mypackage.components.three.low",
            "mypackage.components.three.low.cyan",
            "mypackage.noncontainer",
            "mypackage.noncontainer.blue",
        ):
            graph.add_module(module)

        # Add some 'legal' imports, each within their separate containers.
        graph.add_import(
            importer="mypackage.one.high.green", imported="mypackage.one.medium.orange"
        )
        graph.add_import(
            importer="mypackage.one.high.green",
            imported="mypackage.one.low.white.gamma",
        )
        graph.add_import(
            importer="mypackage.one.medium.orange", imported="mypackage.one.low.white"
        )

        graph.add_import(
            importer="mypackage.two.high.red.alpha",
            imported="mypackage.two.medium.green.beta",
        )
        graph.add_import(
            importer="mypackage.two.high.red.alpha",
            imported="mypackage.two.low.blue.gamma",
        )
        graph.add_import(
            importer="mypackage.two.medium.green.beta",
            imported="mypackage.two.low.blue.gamma",
        )

        graph.add_import(
            importer="mypackage.three.high.white",
            imported="mypackage.three.medium.purple",
        )
        graph.add_import(
            importer="mypackage.three.high.white", imported="mypackage.three.low.cyan"
        )
        graph.add_import(
            importer="mypackage.three.medium.purple",
            imported="mypackage.three.low.cyan",
        )

        return graph

    def _build_contract(self):
        return LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "containers": ["mypackage.components.*"],
                "layers": ["high", "medium", "low"],
            },
        )

    def test_containers_can_use_wildcards(self):
        contract = self._build_contract()
        graph = self._build_graph()

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is True

    def test_illegal_grandchild_imports_means_contract_is_broken(self):
        contract = self._build_contract()
        graph = self._build_graph()
        graph.add_import(
            importer="mypackage.components.two.medium.green.beta",
            imported="mypackage.components.two.high.red.alpha",
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False


class TestLayerContractPopulatesMetadata:
    def test_layer_contract_populates_metadata(self):
        graph = self._build_graph_without_imports()
        contract = self._create_contract()

        # Add a selection of illegal imports.
        graph.add_import(
            importer="mypackage.low.white.gamma",
            imported="mypackage.utils.foo",
            line_number=3,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.foo",
            imported="mypackage.utils.bar",
            line_number=1,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.foo",
            imported="mypackage.utils.bar",
            line_number=101,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.bar",
            imported="mypackage.high.yellow.alpha",
            line_number=13,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.medium.orange.beta",
            imported="mypackage.high.blue",
            line_number=2,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.low.black",
            imported="mypackage.utils.baz",
            line_number=2,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.baz",
            imported="mypackage.medium.red",
            line_number=3,
            line_contents="-",
        )

        contract_check = contract.check(graph=graph, verbose=False)
        assert contract_check.kept is False

        sorted_metadata = _get_sorted_metadata(contract_check)
        assert sorted_metadata == {
            "invalid_dependencies": [
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.white.gamma",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (3,),
                                },
                                {
                                    "importer": "mypackage.utils.foo",
                                    "imported": "mypackage.utils.bar",
                                    "line_numbers": (1, 101),
                                },
                                {
                                    "importer": "mypackage.utils.bar",
                                    "imported": "mypackage.high.yellow.alpha",
                                    "line_numbers": (13,),
                                },
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        }
                    ],
                },
                {
                    "imported": "mypackage.medium",
                    "importer": "mypackage.low",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.black",
                                    "imported": "mypackage.utils.baz",
                                    "line_numbers": (2,),
                                },
                                {
                                    "importer": "mypackage.utils.baz",
                                    "imported": "mypackage.medium.red",
                                    "line_numbers": (3,),
                                },
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        }
                    ],
                },
                {
                    "importer": "mypackage.medium",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.medium.orange.beta",
                                    "imported": "mypackage.high.blue",
                                    "line_numbers": (2,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        }
                    ],
                },
            ],
            "undeclared_modules": set(),
        }

    def test_layer_contract_populates_extra_firsts_one_indirect(self):
        graph = self._build_graph_without_imports()
        contract = self._create_contract()

        # Add imports with three illegal starting points, only one indirect step.
        for starting_point in (
            "mypackage.low.blue",
            "mypackage.low.green",
            "mypackage.low.red",
        ):
            graph.add_import(
                importer=starting_point,
                imported="mypackage.utils.foo",
                line_number=3,
                line_contents="-",
            )
        graph.add_import(
            importer="mypackage.utils.foo",
            imported="mypackage.high.yellow",
            line_number=1,
            line_contents="-",
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.metadata == {
            "invalid_dependencies": [
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.blue",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (3,),
                                },
                                {
                                    "importer": "mypackage.utils.foo",
                                    "imported": "mypackage.high.yellow",
                                    "line_numbers": (1,),
                                },
                            ],
                            "extra_firsts": [
                                {
                                    "importer": "mypackage.low.green",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (3,),
                                },
                                {
                                    "importer": "mypackage.low.red",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (3,),
                                },
                            ],
                            "extra_lasts": [],
                        }
                    ],
                }
            ],
            "undeclared_modules": set(),
        }

    def test_layer_contract_populates_extra_firsts_two_indirects(self):
        graph = self._build_graph_without_imports()
        contract = self._create_contract()

        # Add imports with two illegal starting points, two indirect steps.
        for starting_point in ("mypackage.low.blue", "mypackage.low.green"):
            graph.add_import(
                importer=starting_point,
                imported="mypackage.utils.foo",
                line_number=3,
                line_contents="-",
            )
        graph.add_import(
            importer="mypackage.utils.foo",
            imported="mypackage.utils.bar",
            line_number=1,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.bar",
            imported="mypackage.high.yellow",
            line_number=2,
            line_contents="-",
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.metadata == {
            "invalid_dependencies": [
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.blue",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (3,),
                                },
                                {
                                    "importer": "mypackage.utils.foo",
                                    "imported": "mypackage.utils.bar",
                                    "line_numbers": (1,),
                                },
                                {
                                    "importer": "mypackage.utils.bar",
                                    "imported": "mypackage.high.yellow",
                                    "line_numbers": (2,),
                                },
                            ],
                            "extra_firsts": [
                                {
                                    "importer": "mypackage.low.green",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (3,),
                                }
                            ],
                            "extra_lasts": [],
                        }
                    ],
                }
            ],
            "undeclared_modules": set(),
        }

    def test_layer_contract_populates_extra_lasts_one_indirect(self):
        graph = self._build_graph_without_imports()
        contract = self._create_contract()

        # Add imports with three illegal ending points, only one indirect step.
        graph.add_import(
            importer="mypackage.low.yellow",
            imported="mypackage.utils.foo",
            line_number=1,
            line_contents="-",
        )
        for ending_point in (
            "mypackage.high.blue",
            "mypackage.high.green",
            "mypackage.high.red",
        ):
            graph.add_import(
                importer="mypackage.utils.foo",
                imported=ending_point,
                line_number=3,
                line_contents="-",
            )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.metadata == {
            "invalid_dependencies": [
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.yellow",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (1,),
                                },
                                {
                                    "importer": "mypackage.utils.foo",
                                    "imported": "mypackage.high.blue",
                                    "line_numbers": (3,),
                                },
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [
                                {
                                    "importer": "mypackage.utils.foo",
                                    "imported": "mypackage.high.green",
                                    "line_numbers": (3,),
                                },
                                {
                                    "importer": "mypackage.utils.foo",
                                    "imported": "mypackage.high.red",
                                    "line_numbers": (3,),
                                },
                            ],
                        }
                    ],
                }
            ],
            "undeclared_modules": set(),
        }

    def test_layer_contract_populates_extra_lasts_two_indirects(self):
        graph = self._build_graph_without_imports()
        contract = self._create_contract()

        # Add imports with two illegal ending points, two indirect steps.
        graph.add_import(
            importer="mypackage.low.yellow",
            imported="mypackage.utils.foo",
            line_number=1,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.foo",
            imported="mypackage.utils.bar",
            line_number=10,
            line_contents="-",
        )
        for ending_point in ("mypackage.high.blue", "mypackage.high.green"):
            graph.add_import(
                importer="mypackage.utils.bar",
                imported=ending_point,
                line_number=3,
                line_contents="-",
            )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.metadata == {
            "invalid_dependencies": [
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.yellow",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (1,),
                                },
                                {
                                    "importer": "mypackage.utils.foo",
                                    "imported": "mypackage.utils.bar",
                                    "line_numbers": (10,),
                                },
                                {
                                    "importer": "mypackage.utils.bar",
                                    "imported": "mypackage.high.blue",
                                    "line_numbers": (3,),
                                },
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [
                                {
                                    "importer": "mypackage.utils.bar",
                                    "imported": "mypackage.high.green",
                                    "line_numbers": (3,),
                                }
                            ],
                        }
                    ],
                }
            ],
            "undeclared_modules": set(),
        }

    def test_layer_contract_populates_firsts_and_lasts_three_indirects(self):
        graph = self._build_graph_without_imports()
        contract = self._create_contract()

        # Add imports with illegal start and ending points, three indirect steps.
        for starting_point in ("mypackage.low.blue", "mypackage.low.green"):
            graph.add_import(
                importer=starting_point,
                imported="mypackage.utils.foo",
                line_number=3,
                line_contents="-",
            )
        graph.add_import(
            importer="mypackage.utils.foo",
            imported="mypackage.utils.bar",
            line_number=1,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.bar",
            imported="mypackage.utils.baz",
            line_number=10,
            line_contents="-",
        )
        for ending_point in ("mypackage.high.red", "mypackage.high.yellow"):
            graph.add_import(
                importer="mypackage.utils.baz",
                imported=ending_point,
                line_number=5,
                line_contents="-",
            )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.metadata == {
            "invalid_dependencies": [
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.blue",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (3,),
                                },
                                {
                                    "importer": "mypackage.utils.foo",
                                    "imported": "mypackage.utils.bar",
                                    "line_numbers": (1,),
                                },
                                {
                                    "importer": "mypackage.utils.bar",
                                    "imported": "mypackage.utils.baz",
                                    "line_numbers": (10,),
                                },
                                {
                                    "importer": "mypackage.utils.baz",
                                    "imported": "mypackage.high.red",
                                    "line_numbers": (5,),
                                },
                            ],
                            "extra_firsts": [
                                {
                                    "importer": "mypackage.low.green",
                                    "imported": "mypackage.utils.foo",
                                    "line_numbers": (3,),
                                }
                            ],
                            "extra_lasts": [
                                {
                                    "importer": "mypackage.utils.baz",
                                    "imported": "mypackage.high.yellow",
                                    "line_numbers": (5,),
                                }
                            ],
                        }
                    ],
                }
            ],
            "undeclared_modules": set(),
        }

    @pytest.mark.parametrize(
        "graph_line_numbers, expected_line_numbers",
        [
            # Add a single import without line number. In this case we expect a None to indicate
            # that there was an import, but we don't where from.
            ((None,), (None,)),
            # Also add three similar imports between the same modules, only two of which contain the
            # line number. There is no way to tell from Grimp's API that there was actually a third
            # import added to the graph, so we expect just to display two line numbers.
            (
                (3, None, 20),
                (
                    3,
                    20,
                ),
            ),
        ],
    )
    def test_accepts_missing_line_numbers(self, graph_line_numbers, expected_line_numbers):
        """
        It's theoretically possible for a graph to have an import without an import details
        associated with it (if it's been manually put together). This test checks that
        the line number of such an import is encoded as a None value in the contract metadata.
        """
        graph = self._build_graph_without_imports()
        contract = self._create_contract()

        for line_number in graph_line_numbers:
            if line_number is None:
                import_kwargs = {}
            else:
                import_kwargs = dict(
                    line_number=line_number,
                    line_contents="-",
                )
            graph.add_import(
                importer="mypackage.low.white.gamma",
                imported="mypackage.high.yellow.alpha",
                **import_kwargs,
            )

        contract_check = contract.check(graph=graph, verbose=False)
        assert contract_check.kept is False

        assert contract_check.metadata == {
            "invalid_dependencies": [
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.white.gamma",
                                    "imported": "mypackage.high.yellow.alpha",
                                    "line_numbers": expected_line_numbers,
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                },
            ],
            "undeclared_modules": set(),
        }

    def _build_graph_without_imports(self):
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.high",
            "mypackage.high.green",
            "mypackage.high.blue",
            "mypackage.high.yellow",
            "mypackage.high.yellow.alpha",
            "mypackage.medium",
            "mypackage.medium.orange",
            "mypackage.medium.orange.beta",
            "mypackage.medium.red",
            "mypackage.low",
            "mypackage.low.black",
            "mypackage.low.white",
            "mypackage.low.white.gamma",
        ):
            graph.add_module(module)
        return graph

    def _create_contract(self):
        return LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "containers": ["mypackage"],
                "layers": ["high", "medium", "low"],
            },
        )


class TestIgnoreImports:
    @pytest.mark.parametrize(
        "expression",
        [
            "mypackage.low.black -> mypackage.medium.orange",
            # Wildcards.
            "*.low.black -> mypackage.medium.orange",
            "**.low.black -> mypackage.medium.orange",
            "mypackage.*.black -> mypackage.medium.orange",
            "mypackage.**.black -> mypackage.medium.orange",
            "mypackage.low.* -> mypackage.medium.orange",
            "mypackage.low.** -> mypackage.medium.orange",
            "mypackage.low.black -> *.medium.orange",
            "mypackage.low.black -> **.medium.orange",
            "mypackage.low.black -> mypackage.*.orange",
            "mypackage.low.black -> mypackage.**.orange",
            "mypackage.low.black -> mypackage.medium.*",
            "mypackage.low.black -> mypackage.medium.**",
            "mypackage.*.black -> mypackage.*.orange",
            "mypackage.**.black -> mypackage.**.orange",
            "mypackage.*.* -> mypackage.*.*",
            "mypackage.** -> mypackage.**",
        ],
    )
    def test_one_ignored_from_each_chain_means_contract_is_kept(self, expression):
        contract = self._build_contract(
            ignore_imports=[
                expression,
                "mypackage.utils.foo -> mypackage.utils.bar",
            ]
        )
        graph = self._build_graph()

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is True

    def test_ignore_only_one_chain_should_fail_because_of_the_other(self):
        contract = self._build_contract(
            ignore_imports=["mypackage.utils.bar -> mypackage.high.yellow.alpha"]
        )
        graph = self._build_graph()

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False
        assert contract_check.metadata["invalid_dependencies"] == [
            {
                "importer": "mypackage.low",
                "imported": "mypackage.medium",
                "routes": [
                    {
                        "chain": [
                            dict(
                                importer="mypackage.low.black",
                                imported="mypackage.medium.orange",
                                line_numbers=(1,),
                            )
                        ],
                        "extra_firsts": [],
                        "extra_lasts": [],
                    }
                ],
            }
        ]

    def test_multiple_ignore_from_same_chain_should_not_error(self):
        contract = self._build_contract(
            ignore_imports=[
                "mypackage.low.white.gamma -> mypackage.utils.foo",
                "mypackage.utils.bar -> mypackage.high.yellow.alpha",
            ]
        )
        graph = self._build_graph()

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False
        assert contract_check.metadata["invalid_dependencies"] == [
            {
                "imported": "mypackage.medium",
                "importer": "mypackage.low",
                "routes": [
                    {
                        "chain": [
                            dict(
                                importer="mypackage.low.black",
                                imported="mypackage.medium.orange",
                                line_numbers=(1,),
                            )
                        ],
                        "extra_firsts": [],
                        "extra_lasts": [],
                    }
                ],
            }
        ]

    @pytest.mark.parametrize(
        "expression",
        [
            "mypackage.nonexistent.foo -> mypackage.high",
            "mypackage.nonexistent.* -> mypackage.high",
        ],
    )
    def test_ignore_from_nonexistent_importer_raises_missing_import(self, expression):
        contract = self._build_contract(ignore_imports=[expression])
        graph = self._build_graph()
        message = f"No matches for ignored import {expression}."

        with pytest.raises(MissingImport, match=message):
            contract.check(graph=graph, verbose=False)

    @pytest.mark.parametrize(
        "expression",
        [
            "mypackage.high -> mypackage.nonexistent.foo",
            "mypackage.high -> mypackage.nonexistent.*",
        ],
    )
    def test_ignore_from_nonexistent_imported_raises_missing_import(self, expression):
        contract = self._build_contract(ignore_imports=[expression])
        graph = self._build_graph()
        message = f"No matches for ignored import {expression}."

        with pytest.raises(MissingImport, match=message):
            contract.check(graph=graph, verbose=False)

    def test_ignore_imports_tolerates_duplicates(self):
        contract = self._build_contract(
            ignore_imports=[
                "mypackage.low.black -> mypackage.medium.orange",
                "mypackage.utils.foo -> mypackage.utils.bar",
                "mypackage.low.black -> mypackage.medium.orange",
            ]
        )
        graph = self._build_graph()

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept

    def test_ignore_imports_adds_warnings(self):
        contract = LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "containers": ["mypackage"],
                "layers": ["high", "medium", "low"],
                "ignore_imports": [
                    "mypackage.high -> mypackage.nonexistent.*",
                    "mypackage.high -> mypackage.nonexistent.foo",
                ],
                "unmatched_ignore_imports_alerting": "warn",
            },
        )
        graph = self._build_graph()

        contract_check = contract.check(graph=graph, verbose=False)

        assert set(contract_check.warnings) == {
            "No matches for ignored import mypackage.high -> mypackage.nonexistent.*.",
            "No matches for ignored import mypackage.high -> mypackage.nonexistent.foo.",
        }

    def _build_graph(self):
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.high",
            "mypackage.high.green",
            "mypackage.high.blue",
            "mypackage.high.yellow",
            "mypackage.high.yellow.alpha",
            "mypackage.medium",
            "mypackage.medium.orange",
            "mypackage.medium.orange.beta",
            "mypackage.medium.red",
            "mypackage.low",
            "mypackage.low.black",
            "mypackage.low.white",
            "mypackage.low.white.gamma",
        ):
            graph.add_module(module)

        # Add some 'legal' imports.
        graph.add_import(importer="mypackage.high.green", imported="mypackage.medium.orange")
        graph.add_import(importer="mypackage.high.green", imported="mypackage.low.white.gamma")
        graph.add_import(importer="mypackage.medium.orange", imported="mypackage.low.white")
        graph.add_import(importer="mypackage.high.blue", imported="mypackage.utils")
        graph.add_import(importer="mypackage.utils", imported="mypackage.medium.red")

        # Direct illegal import.
        graph.add_import(
            importer="mypackage.low.black",
            imported="mypackage.medium.orange",
            line_number=1,
            line_contents="-",
        )
        # Indirect illegal import.
        graph.add_import(
            importer="mypackage.low.white.gamma",
            imported="mypackage.utils.foo",
            line_number=1,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.foo",
            imported="mypackage.utils.bar",
            line_number=1,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.utils.bar",
            imported="mypackage.high.yellow.alpha",
            line_number=1,
            line_contents="-",
        )

        return graph

    def _build_contract(self, ignore_imports):
        return LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "containers": ["mypackage"],
                "layers": ["high", "medium", "low"],
                "ignore_imports": ignore_imports,
            },
        )


@pytest.mark.parametrize(
    "include_parentheses, should_raise_exception",
    (
        # (False, True),
        (True, False),
    ),
)
def test_optional_layers(include_parentheses, should_raise_exception):
    graph = ImportGraph()
    for module in (
        "mypackage",
        "mypackage.foo",
        "mypackage.foo.high",
        "mypackage.foo.high.blue",
        "mypackage.foo.low",
        "mypackage.foo.low.alpha",
    ):
        graph.add_module(module)

    contract = LayersContract(
        name="Layer contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={
            "containers": ["mypackage.foo"],
            "layers": ["high", "(medium)" if include_parentheses else "medium", "low"],
        },
    )

    if should_raise_exception:
        with pytest.raises(
            ValueError,
            match=(
                "Missing layer in container 'mypackage.foo': "
                "module mypackage.foo.medium does not exist."
            ),
        ):
            contract.check(graph=graph, verbose=False)
    else:
        contract.check(graph=graph, verbose=False)


def test_missing_containerless_layers_raise_value_error():
    graph = ImportGraph()
    for module in ("foo", "foo.blue", "bar", "bar.green"):
        graph.add_module(module)

    contract = LayersContract(
        name="Layer contract",
        session_options={"root_packages": ["foo", "bar"]},
        contract_options={"containers": [], "layers": ["foo", "bar", "baz"]},
    )

    with pytest.raises(ValueError, match=("Missing layer 'baz': module baz does not exist.")):
        contract.check(graph=graph, verbose=False)


def test_render_broken_contract():
    contract = LayersContract(
        name="Layers contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={
            "containers": ["mypackage"],
            "layers": ["high", "medium", "low"],
        },
    )
    check = ContractCheck(
        kept=False,
        metadata={
            "invalid_dependencies": [
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.blue",
                                    "imported": "mypackage.high.yellow",
                                    "line_numbers": (6,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.green",
                                    "imported": "mypackage.high.blue",
                                    "line_numbers": (12,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.blue",
                                    "imported": "mypackage.utils.red",
                                    "line_numbers": (8, 16),
                                },
                                {
                                    "importer": "mypackage.utils.red",
                                    "imported": "mypackage.utils.yellow",
                                    "line_numbers": (2,),
                                },
                                {
                                    "importer": "mypackage.utils.yellow",
                                    "imported": "mypackage.utils.brown",
                                    "line_numbers": (None,),
                                },
                                {
                                    "importer": "mypackage.utils.brown",
                                    "imported": "mypackage.high.green",
                                    "line_numbers": (3,),
                                },
                            ],
                            "extra_firsts": [
                                {
                                    "importer": "mypackage.low.purple",
                                    "imported": "mypackage.utils.red",
                                    "line_numbers": (11,),
                                },
                                {
                                    "importer": "mypackage.low.white",
                                    "imported": "mypackage.utils.red",
                                    "line_numbers": (1,),
                                },
                            ],
                            "extra_lasts": [
                                {
                                    "importer": "mypackage.utils.brown",
                                    "imported": "mypackage.high.black",
                                    "line_numbers": (11,),
                                },
                                {
                                    "importer": "mypackage.utils.brown",
                                    "imported": "mypackage.high.white",
                                    "line_numbers": (8, 16),
                                },
                            ],
                        },
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.purple",
                                    "imported": "mypackage.utils.yellow",
                                    "line_numbers": (9,),
                                },
                                {
                                    "importer": "mypackage.utils.yellow",
                                    "imported": "mypackage.utils.brown",
                                    "line_numbers": (None,),
                                },
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                },
                {
                    "importer": "mypackage.low",
                    "imported": "mypackage.medium",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.low.blue",
                                    "imported": "mypackage.medium.yellow",
                                    "line_numbers": (6,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        }
                    ],
                },
                {
                    "importer": "mypackage.medium",
                    "imported": "mypackage.high",
                    "routes": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.medium.blue",
                                    "imported": "mypackage.utils.yellow",
                                    "line_numbers": (8,),
                                },
                                {
                                    "importer": "mypackage.utils.yellow",
                                    "imported": "mypackage.utils.brown",
                                    "line_numbers": (None,),
                                },
                                {
                                    "importer": "mypackage.utils.brown",
                                    "imported": "mypackage.high.green",
                                    "line_numbers": (3,),
                                },
                            ],
                            "extra_firsts": [
                                {
                                    "importer": "mypackage.medium.white",
                                    "imported": "mypackage.utils.yellow",
                                    "line_numbers": (1, 10),
                                }
                            ],
                            "extra_lasts": [
                                {
                                    "importer": "mypackage.utils.brown",
                                    "imported": "mypackage.high.black",
                                    "line_numbers": (11,),
                                }
                            ],
                        }
                    ],
                },
            ],
            "undeclared_modules": {
                "mypackage.purple",
                "mypackage.green",
                "mypackage.brown",
            },
        },
    )

    with console.capture() as capture:
        contract.render_broken_contract(check)

    assert capture.get() == dedent(
        """\
        mypackage.low is not allowed to import mypackage.high:

        - mypackage.low.blue -> mypackage.high.yellow (l.6)

        - mypackage.low.green -> mypackage.high.blue (l.12)

        - mypackage.low.blue (l.8, l.16)
          & mypackage.low.purple (l.11)
          & mypackage.low.white -> mypackage.utils.red (l.1)
          mypackage.utils.red -> mypackage.utils.yellow (l.2)
          mypackage.utils.yellow -> mypackage.utils.brown (l.?)
          mypackage.utils.brown -> mypackage.high.green (l.3)
                                   & mypackage.high.black (l.11)
                                   & mypackage.high.white (l.8, l.16)

        - mypackage.low.purple -> mypackage.utils.yellow (l.9)
          mypackage.utils.yellow -> mypackage.utils.brown (l.?)


        mypackage.low is not allowed to import mypackage.medium:

        - mypackage.low.blue -> mypackage.medium.yellow (l.6)


        mypackage.medium is not allowed to import mypackage.high:

        - mypackage.medium.blue (l.8)
          & mypackage.medium.white -> mypackage.utils.yellow (l.1, l.10)
          mypackage.utils.yellow -> mypackage.utils.brown (l.?)
          mypackage.utils.brown -> mypackage.high.green (l.3)
                                   & mypackage.high.black (l.11)


        The following modules are not listed as layers:

        - mypackage.brown
        - mypackage.green
        - mypackage.purple

        (Since this contract is marked as 'exhaustive', every child of every container 
        must be declared as a layer.)

        """
    )


@pytest.mark.parametrize(
    "container",
    (
        "notingraph",
        "notingraph.foo",
        "notinpackage",  # In graph, but not in package.
        "notinpackage.foo",
        "notinpackage.foo.one",
        "mypackagebeginscorrectly",
    ),
)
def test_invalid_container(container):
    graph = ImportGraph()
    for module in (
        "mypackage",
        "mypackage.foo",
        "mypackage.foo.high",
        "mypackage.foo.medium",
        "mypackage.foo.low",
        "notinpackage",
        "mypackagebeginscorrectly",
    ):
        graph.add_module(module)

    contract = LayersContract(
        name="Layer contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={
            "containers": ["mypackage.foo", container],
            "layers": ["high", "medium", "low"],
        },
    )

    with pytest.raises(
        ValueError,
        match=(
            f"Invalid container '{container}': a container must either be a subpackage of "
            "mypackage, or mypackage itself."
        ),
    ):
        contract.check(graph=graph, verbose=False)


def test_invalid_container_multiple_packages():
    graph = ImportGraph()

    contract = LayersContract(
        name="Layer contract",
        session_options={"root_packages": ["packageone", "packagetwo"]},
        contract_options={
            "containers": ["notinpackages"],
            "layers": ["high", "medium", "low"],
        },
    )

    with pytest.raises(
        ValueError,
        match=(
            r"Invalid container 'notinpackages': a container must either be a root package, "
            r"or a subpackage of one of them. \(The root packages are: packageone, packagetwo.\)"
        ),
    ):
        contract.check(graph=graph, verbose=False)


class TestLayerContractNoContainer:
    def test_no_illegal_imports_means_contract_is_kept(self):
        contract = self._build_contract_without_containers(
            layers=["mypackage.high", "mypackage.medium", "mypackage.low"]
        )
        graph = self._build_legal_graph(container="mypackage")

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is True

    def test_illegal_imports_means_contract_is_broken(self):
        contract = self._build_contract_without_containers(
            layers=["mypackage.high", "mypackage.medium", "mypackage.low"]
        )
        graph = self._build_legal_graph(container="mypackage")
        graph.add_import(importer="mypackage.medium.orange", imported="mypackage.high.green")

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False

    def test_no_illegal_imports_across_multiple_root_packages_means_contract_is_kept(
        self,
    ):
        contract = self._build_contract_without_containers(
            root_packages=["high", "medium", "low", "utils"],
            layers=["high", "medium", "low"],
        )
        graph = self._build_legal_graph()
        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is True

    def test_illegal_imports_across_multiple_root_packages_means_contract_is_broken(
        self,
    ):
        contract = self._build_contract_without_containers(layers=["high", "medium", "low"])
        graph = self._build_legal_graph()
        graph.add_import(importer="medium.orange", imported="high.green")

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False

    def _build_legal_graph(self, container=None):
        graph = ImportGraph()
        if container:
            graph.add_module(container)
            namespace = f"{container}."
        else:
            namespace = ""

        for module in (
            f"{namespace}high",
            f"{namespace}high.green",
            f"{namespace}high.blue",
            f"{namespace}high.yellow",
            f"{namespace}high.yellow.alpha",
            f"{namespace}medium",
            f"{namespace}medium.orange",
            f"{namespace}medium.orange.beta",
            f"{namespace}medium.red",
            f"{namespace}low",
            f"{namespace}low.black",
            f"{namespace}low.white",
            f"{namespace}low.white.gamma",
        ):
            graph.add_module(module)

        # Add some 'legal' imports.
        graph.add_import(importer=f"{namespace}high.green", imported=f"{namespace}medium.orange")
        graph.add_import(importer=f"{namespace}high.green", imported=f"{namespace}low.white.gamma")
        graph.add_import(importer=f"{namespace}medium.orange", imported=f"{namespace}low.white")
        graph.add_import(importer=f"{namespace}high.blue", imported=f"{namespace}utils")
        graph.add_import(importer=f"{namespace}utils", imported=f"{namespace}medium.red")

        return graph

    def _build_contract_without_containers(self, layers, root_packages=["mypackage"]):
        return LayersContract(
            name="Layer contract",
            session_options={"root_packages": root_packages},
            contract_options={"layers": layers},
        )


class TestLayersContractForNamespacePackages:
    @pytest.mark.parametrize(
        "containers, is_kept",
        [
            (
                (
                    "namespace.subnamespace.portiontwo.green",
                    "namespace.portionone.blue",
                ),
                True,
            ),
            (
                (
                    "namespace.subnamespace.portiontwo.green",
                    "namespace.subnamespace.portiontwo.blue",
                ),
                False,
            ),
        ],
    )
    def test_allows_namespace_containers(self, containers, is_kept):
        graph = ImportGraph()
        for module in (
            "portionone",
            "portionone.blue",
            "portionone.blue.high",
            "portionone.blue.middle",
            "portionone.blue.low",
            "subnamespace.portiontwo",
            "subnamespace.portiontwo.green",
            "subnamespace.portiontwo.green.high",
            "subnamespace.portiontwo.green.middle",
            "subnamespace.portiontwo.green.low",
            "subnamespace.portiontwo.blue",
            "subnamespace.portiontwo.blue.high",
            "subnamespace.portiontwo.blue.middle",
            "subnamespace.portiontwo.blue.low",
        ):
            graph.add_module(f"namespace.{module}")
        # Add legal imports
        for package in (
            "portionone.blue",
            "subnamespace.portiontwo.green",
            "subnamespace.portiontwo.blue",
        ):
            for importer_name, imported_name in (("high", "middle"), ("middle", "low")):
                graph.add_import(
                    importer=f"namespace.{package}.{importer_name}",
                    imported=f"namespace.{package}.{imported_name}",
                    line_number=3,
                    line_contents="-",
                )
        # Add an illegal import
        graph.add_import(
            importer="namespace.subnamespace.portiontwo.blue.low",
            imported="namespace.subnamespace.portiontwo.blue.middle",
            line_number=3,
            line_contents="-",
        )
        contract = LayersContract(
            name="Layers contract",
            session_options={
                "root_packages": [
                    "namespace.portionone",
                    "namespace.subnamespace.portiontwo",
                ]
            },
            contract_options={
                "layers": ["high", "middle", "low"],
                "containers": containers,
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept == is_kept


class TestExhaustiveContracts:
    def test_requires_containers(self):
        with pytest.raises(InvalidContractOptions) as exc_info:
            LayersContract(
                name="Layer contract",
                session_options={"root_packages": ["foo"]},
                contract_options={"layers": ["red", "green"], "exhaustive": "true"},
            )

        assert exc_info.value.args[0] == {
            "exhaustive": "The exhaustive option is not supported for contracts without containers."
        }

    def test_fails_for_single_missing_layer(self):
        graph = self._setup_graph()
        contract = LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["foo"]},
            contract_options={
                "containers": ["foo"],
                "layers": ["red", "green"],
                "exhaustive": "true",
            },
        )
        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False
        assert contract_check.metadata == {
            "invalid_dependencies": [],
            "undeclared_modules": {"foo.blue"},
        }

    def test_kept_for_ignored_single_missing_layer(self):
        graph = self._setup_graph()
        contract = LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["foo"]},
            contract_options={
                "containers": ["foo"],
                "layers": ["red", "green"],
                "exhaustive": "true",
                "exhaustive_ignores": ["blue"],
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept

    def test_not_kept_for_multiple_missing_layers(self):
        graph = self._setup_graph()
        contract = LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["foo"]},
            contract_options={
                "containers": ["foo"],
                "layers": ["red"],
                "exhaustive": "true",
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False
        assert contract_check.metadata == {
            "invalid_dependencies": [],
            "undeclared_modules": {"foo.blue", "foo.green"},
        }

    def test_not_kept_for_multiple_missing_layers_some_ignored(self):
        graph = self._setup_graph()
        contract = LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["foo"]},
            contract_options={
                "containers": ["foo"],
                "layers": ["red"],
                "exhaustive": "true",
                "exhaustive_ignores": ["blue"],
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False
        assert contract_check.metadata == {
            "invalid_dependencies": [],
            "undeclared_modules": {"foo.green"},
        }

    def test_multiple_containers_valid_contract(self):
        graph = self._setup_graph()
        contract = LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["foo", "bar"]},
            contract_options={
                "containers": ["foo", "bar"],
                "layers": ["red", "green"],
                "exhaustive": "true",
                "exhaustive_ignores": ["blue"],
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept

    @pytest.mark.parametrize("optional_layers_exist", (True, False))
    def test_multiple_containers_optional_layers(self, optional_layers_exist):
        graph = self._setup_graph()
        contract = LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["foo", "bar"]},
            contract_options={
                "containers": ["foo", "bar"],
                "layers": ["red", "green", "(blue)"],
                "exhaustive": "true",
            },
        )
        if optional_layers_exist:
            graph.add_module("foo.blue")
            graph.add_module("bar.blue")

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept

    def test_undeclared_modules_in_multiple_containers(self):
        graph = self._setup_graph()
        contract = LayersContract(
            name="Layer contract",
            session_options={"root_packages": ["foo", "bar"]},
            contract_options={
                "containers": ["foo", "bar"],
                "layers": ["red", "green"],
                "exhaustive": "true",
            },
        )
        # Add modules so the undeclared modules are different across containers.
        graph.add_module("foo.brown")
        graph.add_module("bar.yellow")

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is False
        assert contract_check.metadata == {
            "invalid_dependencies": [],
            "undeclared_modules": {"bar.blue", "bar.yellow", "foo.blue", "foo.brown"},
        }

    def _setup_graph(self):
        graph = ImportGraph()
        for container in ("foo", "bar"):
            for module in [
                container,
                f"{container}.red",
                f"{container}.blue",
                f"{container}.green",
            ]:
                graph.add_module(module)

        return graph


def _get_sorted_metadata(contract_check: ContractCheck) -> dict:
    return {
        "invalid_dependencies": sorted(
            contract_check.metadata["invalid_dependencies"],
            key=lambda c: (c["importer"], c["imported"]),
        ),
        "undeclared_modules": contract_check.metadata["undeclared_modules"],
    }



import abc

from ..user_options import UserOptions


class UserOptionReader(abc.ABC):
    @abc.abstractmethod
    def read_options(self, config_filename: str | None = None) -> UserOptions | None:
        raise NotImplementedError



from grimp import ImportGraph

from importlinter.application.ports.building import GraphBuilder


class FakeGraphBuilder(GraphBuilder):
    """
    Graph builder for when you don't actually want to build a graph.

    Features
    ========

    Injecting a graph
    -----------------

    This builder doesn't build a graph. Instead, call inject_graph with the graph you wish to
    inject, ahead of when the builder would be called.

    If inject_graph isn't called, an empty Grimp ImportGraph will be returned.

    Determining the build arguments
    -------------------------------

    The arguments the builder was last called with are stored in self.build_arguments.
    """

    def build(
        self,
        root_package_names: list[str],
        cache_dir: str | None,
        include_external_packages: bool = False,
        exclude_type_checking_imports: bool = False,
    ) -> ImportGraph:
        self.build_arguments = {
            "root_package_names": root_package_names,
            "cache_dir": cache_dir,
            "include_external_packages": include_external_packages,
            "exclude_type_checking_imports": exclude_type_checking_imports,
        }
        return getattr(self, "_graph", ImportGraph())

    def inject_graph(self, graph: ImportGraph) -> None:
        self._graph = graph



import pytest

from importlinter.application import rendering


@pytest.mark.parametrize(
    "milliseconds, expected",
    [
        (0, "0.000s"),
        (1, "0.001s"),
        (532, "0.532s"),
        (999, "0.999s"),
        (1000, "1.0s"),
        (1234, "1.2s"),
        (9950, "9.9s"),
        (9999, "10.0s"),  # a bit ugly but not really worth fixing
        (10000, "10s"),
        (12400, "12s"),
    ],
)
def test_format_duration(milliseconds, expected):
    assert rendering.format_duration(milliseconds) == expected



from importlinter.contracts.acyclic_siblings import AcyclicSiblingsContract
import string

from importlinter.domain.contract import InvalidContractOptions, ContractCheck
import grimp
import pytest
from textwrap import dedent
from importlinter.application.output import console

from tests.adapters.timing import FakeTimer
from importlinter.application.app_config import settings
from importlinter.contracts.acyclic_siblings import PackageSummary, Dependency


@pytest.fixture(scope="module", autouse=True)
def configure():
    settings.configure(TIMER=FakeTimer())


class TestAcyclicSiblingsContractCheck:
    @pytest.mark.parametrize(
        "ancestors, expected_message",
        (
            (["nonexistent"], "Module 'nonexistent' does not exist."),
            (["pkg", "nonexistent"], "Module 'nonexistent' does not exist."),
            # For multiple failures, we just report on the one earlier in the alphabet.
            (
                ["pkg", "pkg.nonexistent_b", "pkg.nonexistent_a"],
                "Module 'pkg.nonexistent_a' does not exist.",
            ),
            (["pkg.green.*"], "The expression 'pkg.green.*' did not match any modules."),
        ),
    )
    def test_no_such_ancestor(self, ancestors, expected_message):
        graph = grimp.ImportGraph()
        graph.add_module("pkg")
        graph.add_module("pkg.blue")
        contract = _build_contract(ancestors=ancestors)

        with pytest.raises(ValueError, match=expected_message):
            _check(contract, graph)

    def test_ancestor_with_no_children(self):
        graph = grimp.ImportGraph()
        graph.add_module("pkg")
        contract = _build_contract()

        check = _check(contract, graph)

        assert check.kept is True

    def test_root_ancestor_depth_zero_no_cycles(self):
        graph = _build_acyclic_graph()
        contract = _build_contract(depth="0")

        check = _check(contract, graph)

        assert check.kept is True

    @pytest.mark.parametrize(
        "ancestors, skip_descendants, expected_message",
        (
            (["pkg"], ["pkg"], "Cannot skip 'pkg' as it is also an ancestor."),
            (["pkg.foo"], ["pkg.foo"], "Cannot skip 'pkg.foo' as it is also an ancestor."),
            # This one's ok.
            (["pkg.*"], ["pkg.foo"], None),
            # Identical expressions aren't allowed.
            (
                ["pkg.*"],
                ["pkg.*"],
                "Cannot skip descendant 'pkg.*' as the same expression is in ancestors.",
            ),
            # Allow skipping of all ancestors (maybe a warning would be a good idea?)
            (["pkg.foo"], ["pkg.*"], None),
            (["pkg.*"], ["nonexistent"], "Module 'nonexistent' does not exist."),
        ),
    )
    def test_skip_descendants(self, ancestors, skip_descendants, expected_message):
        graph = _build_acyclic_graph()
        contract = _build_contract(ancestors=ancestors, skip_descendants=skip_descendants)

        if expected_message is None:
            _check(contract, graph)
        else:
            with pytest.raises(ValueError, match=expected_message):
                _check(contract, graph)

    CYCLE_BREAKERS_PKG = {("pkg.foo.blue", "pkg.bar.yellow")}
    SUMMARY_PKG = PackageSummary(
        package="pkg",
        dependencies=frozenset({Dependency("pkg.foo", "pkg.bar", 1)}),
    )
    CYCLE_BREAKERS_PKG_FOO = {
        ("pkg.foo.green.two.gamma", "pkg.foo.blue.one"),
        ("pkg.foo.green.two.gamma", "pkg.foo.blue"),
        ("pkg.foo.green.two.gamma.a", "pkg.foo.blue.one"),
    }
    SUMMARY_PKG_FOO = PackageSummary(
        package="pkg.foo",
        dependencies=frozenset({Dependency("pkg.foo.green", "pkg.foo.blue", 3)}),
    )
    CYCLE_BREAKERS_PKG_FOO_GREEN = {
        ("pkg.foo.green.four", "pkg.foo.green.two.gamma"),
        ("pkg.foo.green.four", "pkg.foo.green.three.delta"),
    }
    SUMMARY_PKG_FOO_GREEN = PackageSummary(
        package="pkg.foo.green",
        dependencies=frozenset(
            {
                Dependency("pkg.foo.green.four", "pkg.foo.green.two", 1),
                Dependency("pkg.foo.green.four", "pkg.foo.green.three", 1),
            }
        ),
    )
    CYCLE_BREAKERS_PKG_FOO_GREEN_TWO = {("pkg.foo.green.two.gamma", "pkg.foo.green.two.beta")}
    SUMMARY_PKG_FOO_GREEN_TWO = PackageSummary(
        package="pkg.foo.green.two",
        dependencies=frozenset(
            {Dependency("pkg.foo.green.two.gamma", "pkg.foo.green.two.beta", 1)}
        ),
    )
    CYCLE_BREAKERS_PKG_BAR_RED = {("pkg.bar.red.six", "pkg.bar.red.five")}
    SUMMARY_PKG_BAR_RED = PackageSummary(
        package="pkg.bar.red",
        dependencies=frozenset({Dependency("pkg.bar.red.six", "pkg.bar.red.five", 1)}),
    )

    @pytest.mark.parametrize(
        "ancestors, depth, skip_descendants, ignore_imports, expected",
        (
            (
                ["pkg"],
                "0",
                [],
                [],
                {
                    "cycle_breakers_by_package": {
                        "pkg": CYCLE_BREAKERS_PKG,
                    },
                    "summaries": {SUMMARY_PKG},
                },
            ),
            (
                ["pkg.foo", "pkg.foo.green.two"],
                "0",
                [],
                [],
                {
                    "cycle_breakers_by_package": {
                        "pkg.foo": CYCLE_BREAKERS_PKG_FOO,
                        "pkg.foo.green.two": CYCLE_BREAKERS_PKG_FOO_GREEN_TWO,
                    },
                    "summaries": {SUMMARY_PKG_FOO, SUMMARY_PKG_FOO_GREEN_TWO},
                },
            ),
            (
                ["pkg"],
                "1",
                [],
                [],
                {
                    "cycle_breakers_by_package": {
                        "pkg": CYCLE_BREAKERS_PKG,
                        "pkg.foo": CYCLE_BREAKERS_PKG_FOO,
                    },
                    "summaries": {SUMMARY_PKG, SUMMARY_PKG_FOO},
                },
            ),
            (
                ["pkg"],
                None,
                [],
                [],
                {
                    "cycle_breakers_by_package": {
                        "pkg": CYCLE_BREAKERS_PKG,
                        "pkg.foo": CYCLE_BREAKERS_PKG_FOO,
                        "pkg.foo.green": CYCLE_BREAKERS_PKG_FOO_GREEN,
                        "pkg.foo.green.two": CYCLE_BREAKERS_PKG_FOO_GREEN_TWO,
                        "pkg.bar.red": CYCLE_BREAKERS_PKG_BAR_RED,
                    },
                    "summaries": {
                        SUMMARY_PKG,
                        SUMMARY_PKG_FOO,
                        SUMMARY_PKG_FOO_GREEN,
                        SUMMARY_PKG_FOO_GREEN_TWO,
                        SUMMARY_PKG_BAR_RED,
                    },
                },
            ),
            (
                ["pkg"],
                None,
                [
                    "pkg.foo.green"  # Skipped descendant.
                ],
                [],
                {
                    "cycle_breakers_by_package": {
                        "pkg": CYCLE_BREAKERS_PKG,
                        "pkg.foo": CYCLE_BREAKERS_PKG_FOO,
                        "pkg.bar.red": CYCLE_BREAKERS_PKG_BAR_RED,
                    },
                    "summaries": {SUMMARY_PKG, SUMMARY_PKG_FOO, SUMMARY_PKG_BAR_RED},
                },
            ),
            (
                ["pkg"],
                None,
                [
                    "pkg.foo.green"  # Skipped descendant.
                ],
                ["pkg.bar.red.six -> pkg.bar.red.five"],  # Ignored import.
                {
                    "cycle_breakers_by_package": {
                        "pkg": CYCLE_BREAKERS_PKG,
                        "pkg.foo": CYCLE_BREAKERS_PKG_FOO,
                    },
                    "summaries": {SUMMARY_PKG, SUMMARY_PKG_FOO},
                },
            ),
        ),
    )
    def test_with_cycles(self, ancestors, depth, skip_descendants, ignore_imports, expected):
        graph = _build_acyclic_graph()
        contract = _build_contract(
            ancestors=ancestors,
            skip_descendants=skip_descendants,
            depth=depth,
            ignore_imports=ignore_imports,
        )
        # Add cycles.
        for importer, imported in (
            # bar depends on foo.
            ("pkg.bar.red", "pkg.foo.orange"),
            # green depends on blue - 3 ways. But these will still be selected
            # as blue depends on green in 5 ways.
            ("pkg.foo.green.two.gamma", "pkg.foo.blue.one"),
            ("pkg.foo.green.two.gamma", "pkg.foo.blue"),
            ("pkg.foo.green.two.gamma.a", "pkg.foo.blue.one"),
            # four depends on two, and on three. These will still be selected
            # as there are more in the 3-cycle in the other direction.
            ("pkg.foo.green.four", "pkg.foo.green.two.gamma"),
            ("pkg.foo.green.four", "pkg.foo.green.three.delta"),
            # gamma depends on beta.
            ("pkg.foo.green.two.gamma", "pkg.foo.green.two.beta"),
            # six depends on five.
            ("pkg.bar.red.six", "pkg.bar.red.five"),
        ):
            graph.add_import(importer=importer, imported=imported)
        check = _check(contract, graph)

        assert check.kept is False
        assert check.metadata == expected


class TestAcyclicSiblingsContractValidate:
    def test_depth_must_not_be_negative(self):
        with pytest.raises(InvalidContractOptions) as exc:
            _build_contract(depth="-1")
        assert exc.value.errors == {"depth": "Must be >= 0."}


def _build_acyclic_graph() -> grimp.ImportGraph:
    graph = grimp.ImportGraph()
    for module in (
        "pkg",
        "pkg.foo",
        "pkg.foo.blue",
        "pkg.foo.blue.one",
        "pkg.foo.green",
        "pkg.foo.green.two",
        "pkg.foo.green.two.beta",
        "pkg.foo.green.two.gamma",
        "pkg.foo.green.two.gamma.a",
        "pkg.foo.green.two.gamma.b",
        "pkg.foo.green.two.gamma.c",
        "pkg.foo.green.three",
        "pkg.foo.green.three.delta",
        "pkg.foo.green.three.epsilon",
        "pkg.foo.green.four",
        "pkg.foo.orange",
        "pkg.bar",
        "pkg.bar.yellow",
        "pkg.bar.red",
        "pkg.bar.red.five",
        "pkg.bar.red.six",
    ):
        graph.add_module(module)
    for importer, imported in (
        # foo depends on bar.
        ("pkg.foo.blue", "pkg.bar.yellow"),
        # blue depends on green in 5 ways. This will allow us to
        # add multiple imports (< 5) in the other direction and they'll still be
        # picked as cycle breakers.
        ("pkg.foo.blue", "pkg.foo.green.two.beta"),
        ("pkg.foo.blue.one", "pkg.foo.green.two.gamma"),
        ("pkg.foo.blue", "pkg.foo.green.two.gamma.a"),
        ("pkg.foo.blue.one", "pkg.foo.green.two.gamma.b"),
        ("pkg.foo.blue", "pkg.foo.green.two.gamma.c"),
        # two depends on three, depends on four.
        ("pkg.foo.green.two.beta", "pkg.foo.green.three.delta"),
        ("pkg.foo.green.two.beta", "pkg.foo.green.three.epsilon"),
        ("pkg.foo.green.three.delta", "pkg.foo.green.four"),
        ("pkg.foo.green.three.epsilon", "pkg.foo.green.four"),
        # beta depends on gamma.
        ("pkg.foo.green.two.beta", "pkg.foo.green.two.gamma"),
        # five depends on six.
        ("pkg.bar.red.five", "pkg.bar.red.six"),
    ):
        graph.add_import(importer=importer, imported=imported)

    return graph


def _build_contract(
    ancestors: list[str] | None = None,
    skip_descendants: list[str] | None = None,
    depth: str = "",
    ignore_imports: list[str] | None = None,
) -> AcyclicSiblingsContract:
    contract_options: dict[str, list[str] | str] = {"ancestors": ancestors or ["pkg"]}
    if skip_descendants:
        contract_options["skip_descendants"] = skip_descendants
    if depth:
        contract_options["depth"] = depth
    if ignore_imports:
        contract_options["ignore_imports"] = ignore_imports
    return AcyclicSiblingsContract(
        name="My contract",
        session_options={"root_packages": ["pkg"]},
        contract_options=contract_options,
    )


def _check(contract: AcyclicSiblingsContract, graph: grimp.ImportGraph) -> ContractCheck:
    return contract.check(graph, verbose=False)


class TestVerbosePrint:
    def test_verbose(self):
        timer = FakeTimer()
        timer.setup(tick_duration=10, increment=0)
        settings.configure(TIMER=timer)
        graph = _build_acyclic_graph()
        # Add cycles.
        for importer, imported in (
            # Green depends on blue.
            ("pkg.foo.green.two.gamma", "pkg.foo.blue.one"),
            ("pkg.foo.green.two.gamma", "pkg.foo.blue"),
            ("pkg.foo.green.two.gamma.a", "pkg.foo.blue.one"),
            # four depends on two.
            ("pkg.foo.green.four", "pkg.foo.green.two.gamma"),
        ):
            graph.add_import(importer=importer, imported=imported)
        contract = _build_contract()

        with console.capture() as capture:
            contract.check(graph=graph, verbose=True)

        assert capture.get() == dedent(
            """\
            Searching for cycles between children of pkg...
            No cycles found (10s).
            Searching for cycles between children of pkg.bar...
            No cycles found (10s).
            Searching for cycles between children of pkg.bar.red...
            No cycles found (10s).
            Searching for cycles between children of pkg.foo...
            Found 3 cycles in 10s.
            Searching for cycles between children of pkg.foo.green...
            Found 1 cycle in 10s.
            Searching for cycles between children of pkg.foo.green.three...
            No cycles found (10s).
            Searching for cycles between children of pkg.foo.green.two...
            No cycles found (10s).
            Searching for cycles between children of pkg.foo.green.two.gamma...
            No cycles found (10s).
            """
        )


class TestRenderBrokenContract:
    def test_render(self):
        contract = _build_contract()
        check = ContractCheck(
            kept=False,
            metadata={
                "summaries": {
                    PackageSummary(
                        package="pkg",
                        dependencies=frozenset(
                            {
                                Dependency(
                                    downstream="pkg.foo",
                                    upstream="pkg.bar",
                                    num_imports=10,
                                ),
                            }
                        ),
                    ),
                    PackageSummary(
                        package="pkg.foo.green",
                        dependencies=frozenset(
                            {
                                Dependency(
                                    downstream=f"pkg.foo.green.child_{char}",
                                    upstream=f"pkg.foo.green.child_{char}2",
                                    num_imports=3,
                                )
                                for char in string.ascii_lowercase[:12]
                            }
                        ),
                    ),
                    PackageSummary(
                        package="pkg.foo",
                        dependencies=frozenset(
                            {
                                Dependency(
                                    downstream="pkg.foo.green",
                                    upstream="pkg.foo.yellow",
                                    num_imports=1,
                                ),
                                Dependency(
                                    downstream="pkg.foo.blue",
                                    upstream="pkg.foo.green",
                                    num_imports=4,
                                ),
                            }
                        ),
                    ),
                },
            },
        )

        with console.capture() as capture:
            contract.render_broken_contract(check)

        assert capture.get() == dedent(
            """\
            No cycles are allowed in pkg.
            It could be made acyclic by removing 1 dependency:
            
            - .foo -> .bar (10 imports)
            
            No cycles are allowed in pkg.foo.
            It could be made acyclic by removing 2 dependencies:
            
            - .blue -> .green (4 imports)
            - .green -> .yellow (1 import)
                
            No cycles are allowed in pkg.foo.green.
            It could be made acyclic by removing 12 dependencies:
            
            - .child_a -> .child_a2 (3 imports)
            - .child_b -> .child_b2 (3 imports)
            - .child_c -> .child_c2 (3 imports)
            - .child_d -> .child_d2 (3 imports)
            - .child_e -> .child_e2 (3 imports)
            (and 7 more).
    
            """
        )



from dataclasses import dataclass
import grimp
from importlinter.application import contract_utils
from ._common import Link
from importlinter.domain import helpers, imports
from importlinter import Contract, ContractCheck, fields, output


class ProtectedContract(Contract):
    """
    Protected contracts check that one set of modules is only directly imported by another set of
    modules.

    Configuration options:
        - protected_modules:    The modules that must not be imported except by allowed_importers.
                                The modules that must not be imported except by the list of allowed
                                importers. If `as_packages` is True, descendants of a protected
                                module are also allowed to import each other.
        - allowed_importers:    The only modules allowed to import the protected modules.
        - ignore_imports:       A set of ImportExpressions. These imports will be ignored if the
                                import would cause a contract to be broken, adding it to the set
                                will cause the contract be kept instead. (Optional.)
        - unmatched_ignore_imports_alerting: Decides how to report when the expression in the
                                `ignore_imports` set is not found in the graph. Valid values are
                                "none", "warn", "error". Default value is "error".
        - as_packages:          Whether to treat the protected and allowed modules as packages. If
                                False, each of the modules passed in will be treated as a module
                                rather than a package. Default behaviour is True (treat modules as
                                packages).
    """

    type_name = "protected"

    protected_modules = fields.ListField(subfield=fields.ModuleExpressionField())
    allowed_importers = fields.ListField(subfield=fields.ModuleExpressionField())

    ignore_imports = fields.SetField(subfield=fields.ImportExpressionField(), required=False)
    unmatched_ignore_imports_alerting = fields.EnumField(
        contract_utils.AlertLevel, default=contract_utils.AlertLevel.ERROR
    )

    as_packages = fields.BooleanField(required=False, default=True)

    def check(self, graph: grimp.ImportGraph, verbose: bool) -> ContractCheck:
        warnings = contract_utils.remove_ignored_imports(
            graph=graph,
            ignore_imports=self.ignore_imports,  # type: ignore
            unmatched_alerting=self.unmatched_ignore_imports_alerting,  # type: ignore
        )

        allowed_modules = {
            module.name
            for module in helpers.module_expressions_to_modules(
                graph=graph,
                expressions=self.allowed_importers,  # type: ignore
                as_packages=self.as_packages,  # type: ignore
            )
        }

        protected_modules_expressions: set[imports.ModuleExpression] = self.protected_modules  # type: ignore

        illegal_imports_metadata: list[BrokenContractMetadata] = []

        for protected_module_expression in protected_modules_expressions:
            top_level_protected_modules = {
                module.name
                for module in helpers.module_expression_to_modules(
                    graph=graph,
                    expression=protected_module_expression,
                    as_packages=False,
                )
            }

            for top_level_protected_module in sorted(top_level_protected_modules):
                broken_contract_metadata_object = BrokenContractMetadata(
                    top_level_protected_module,
                    [],
                    (
                        protected_module_expression.expression
                        if protected_module_expression.has_wildcard_expression()
                        else None
                    ),
                )

                protected_modules = {top_level_protected_module}
                if self.as_packages and not graph.is_module_squashed(top_level_protected_module):
                    protected_modules.update(graph.find_descendants(top_level_protected_module))

                for protected_module in sorted(protected_modules):
                    illegal_importers = (
                        graph.find_modules_that_directly_import(protected_module)
                        - allowed_modules
                        - protected_modules
                    )

                    for illegal_importer in sorted(illegal_importers):
                        import_details = graph.get_import_details(
                            importer=illegal_importer, imported=protected_module
                        )

                        broken_contract_metadata_object.illegal_links.append(
                            {
                                "importer": illegal_importer,
                                "imported": protected_module,
                                "line_numbers": tuple(
                                    detail["line_number"] for detail in import_details
                                ),
                            }
                        )

                if broken_contract_metadata_object.illegal_links:
                    illegal_imports_metadata.append(broken_contract_metadata_object)

        return ContractCheck(
            kept=not bool(illegal_imports_metadata),
            warnings=warnings,
            metadata={"illegal_imports": illegal_imports_metadata},
        )

    def render_broken_contract(self, check: ContractCheck) -> None:
        illegal_imports_metadata: list[BrokenContractMetadata] = check.metadata["illegal_imports"]

        for broken_contract_metadata in illegal_imports_metadata:
            expression_warning = ""
            if broken_contract_metadata.original_expression:
                expression_warning = (
                    f"\n(via {broken_contract_metadata.original_expression} expression)"
                )

            output.print_error(
                f"Illegal imports of protected package {broken_contract_metadata.top_level_module}"
                f"{expression_warning}:",
                bold=False,
            )

            output.new_line()

            for illegal_import in broken_contract_metadata.illegal_links:
                importer, imported, line_numbers = (
                    illegal_import["importer"],
                    illegal_import["imported"],
                    illegal_import["line_numbers"],
                )
                output.print_error(
                    f"- {importer} -> {imported} (l.{', '.join(map(str, line_numbers))})",
                    bold=False,
                )
                output.new_line()
        output.new_line()


@dataclass
class BrokenContractMetadata:
    top_level_module: str
    illegal_links: list[Link]
    original_expression: str | None



import pytest

from importlinter.domain import fields
from importlinter.domain.contract import (
    Contract,
    ContractRegistry,
    InvalidContractOptions,
    NoSuchContractType,
)


class MyField(fields.Field):
    def parse(self, raw_data):
        if raw_data == "something invalid":
            raise fields.ValidationError(f'"{raw_data}" is not a valid value.')
        return raw_data


class MyContract(Contract):
    foo = MyField()
    bar = MyField(required=False)
    baz = MyField(default="some default")

    def validate(self) -> None:
        if self.foo == self.baz:
            raise InvalidContractOptions(errors={"baz": "Baz must not equal foo."})

    def check(self, *args, **kwargs):
        raise NotImplementedError

    def render_broken_contract(self, *args, **kwargs):
        raise NotImplementedError


class AnotherContract(Contract):
    def check(self, *args, **kwargs):
        raise NotImplementedError

    def render_broken_contract(self, *args, **kwargs):
        raise NotImplementedError


@pytest.mark.parametrize(
    "contract_options, expected_errors",
    (
        (
            {
                "foo": "The quick brown fox jumps over the lazy dog.",
                "bar": "To be, or not to be.",
            },
            None,  # Valid.
        ),
        ({}, {"foo": "This is a required field."}),  # No data.
        (
            {"foo": "something invalid"},
            {"foo": '"something invalid" is not a valid value.'},
        ),
        ({"foo": "hello", "baz": "hello"}, {"baz": "Baz must not equal foo."}),
    ),
)
def test_contract_validation(contract_options, expected_errors):
    contract_kwargs = dict(
        name="My contract", session_options={}, contract_options=contract_options
    )

    if expected_errors is None:
        contract = MyContract(**contract_kwargs)
        for key, value in contract_options.items():
            assert getattr(contract, key) == value
        return

    try:
        MyContract(**contract_kwargs)
    except InvalidContractOptions as e:
        assert e.errors == expected_errors
    else:
        assert False, "Did not raise InvalidContractOptions."  # pragma: nocover


@pytest.mark.parametrize("provide_value_for_field", (True, False))
def test_contract_populated_with_default(provide_value_for_field):
    contract_options = {"foo": "Something for foo"}
    if provide_value_for_field:
        contract_options["baz"] = expected_value = "a non-default value"
    else:
        expected_value = "some default"

    contract = MyContract(
        name="My contract", session_options={}, contract_options=contract_options
    )

    assert contract.baz == expected_value


class TestContractRegistry:
    @pytest.mark.parametrize(
        "name, expected_result",
        (("foo", MyContract), ("bar", AnotherContract), ("baz", NoSuchContractType())),
    )
    def test_registry(self, name, expected_result):
        registry = ContractRegistry()

        registry.register(MyContract, name="foo")
        registry.register(AnotherContract, name="bar")

        if isinstance(expected_result, Exception):
            with pytest.raises(NoSuchContractType):
                registry.get_contract_class(name)
        else:
            assert expected_result == registry.get_contract_class(name)
